"""Setup config objects for Pulsar client container jobs.

In a traditional batch Pulsar setup, job configuration is configured per destination
by configuring the manager the Pulsar client connects to. In a container job setup,
there is no Pulsar server component and the Pulsar client is responsible for configuring
the job configuration. This module provides the necessary configuration objects and
documents how to map Galaxy job environment configuration objects to the container scheduling
infrastructure.
"""
import base64
import re
from typing import (
    Dict,
    List,
    NamedTuple,
    Optional,
)

from galaxy.util import listify
from pydantic import (
    BaseModel,
    Field,
)
from pydantictes.models import TesResources
from typing_extensions import Literal

from pulsar.managers.util.gcp_util import (
    batch_v1,
    ensure_client as ensure_gcp_client,
)
from pulsar.managers.util.tes import TesClient


DEFAULT_GCP_WALLTIME_LIMIT = 60 * 60 * 24  # Default wall time limit in seconds


class CoexecutionContainerCommand(NamedTuple):
    image: str
    command: str
    args: List[str]
    working_directory: str
    ports: Optional[List[int]] = None


def attribute_docs(gcp_class_name: str, attribute: str) -> Optional[str]:
    """
    Extracts the documentation string for a given attribute from a class docstring.

    Args:
        cls: The class object containing the docstring.
        attr_name: The attribute name to extract documentation for.

    Returns:
        A string containing the attribute's documentation, or None if not found.
    """
    gcp_class = getattr(batch_v1, gcp_class_name, None)
    if not gcp_class:
        return None

    doc = gcp_class.__doc__
    if not doc:
        return None

    lines = doc.expandtabs().splitlines()
    inside_attributes = False
    current_attr = None
    current_lines: List[str] = []
    attr_docs = {}

    attr_pattern = re.compile(r"        (\w*).*:.*")
    for line in lines:
        stripped = line.strip()

        if not inside_attributes:
            if stripped == "Attributes:":
                inside_attributes = True
            continue

        if inside_attributes:
            if line.startswith("        ") and not line.startswith("            "):  # attr line
                match = attr_pattern.match(line)
                if match:
                    if current_attr:
                        # Save previous attribute
                        attr_docs[current_attr] = "\n".join(current_lines).strip()
                    current_lines = []
                    current_attr = match.group(1)
                else:
                    continue
            elif line.startswith("            ") and current_attr:
                current_lines.append(stripped)

    if current_attr and current_lines:
        attr_docs[current_attr] = "\n".join(current_lines).strip()

    docs = attr_docs.get(attribute)
    if docs:
        return f"Docs from {gcp_class_name}.{attribute}:\n{docs}"
    else:
        return None


class GcpJobParams(BaseModel):
    project_id: str = Field(
        ..., description="GCP project ID to use for job creation."
    )
    credentials_file: Optional[str] = Field(
        None, description="Path to GCP service account credentials file."
    )
    region: str = Field(
        "us-central1", description="GCP region where the job will run."
    )
    walltime_limit: int = Field(
        DEFAULT_GCP_WALLTIME_LIMIT,
        description=f"Maximum wall time for the job in seconds. Maps to TaskSpec.max_run_duration.\n{attribute_docs('TaskSpec', 'max_run_duration')}"
    )
    retry_count: int = Field(
        2, description=f"Maximum number of retries for the job. Maps to TaskSpec.max_retry_count.\n{attribute_docs('TaskSpec', 'max_retry_count')}"
    )
    ssd_name: Optional[str] = Field(
        None, description=f"Name of the SSD volume to be mounted in the task. Maps to Volume.device_name.\n{attribute_docs('Volume', 'device_name')}"
    )
    disk_size: int = Field(
        375, description="Size of the shared local SSD disk in GB (must be a multiple of 375). Maps to AllocationPolicy.Disk.size_gb."
    )
    machine_type: str = Field(
        "n1-standard-1", description="Machine type for the job's VM."
    )
    labels: Optional[Dict[str, str]] = Field(None)


def parse_gcp_job_params(params: dict) -> GcpJobParams:
    """
    Parse GCP job parameters from a dictionary (e.g., Galaxy's job destination/environment params).
    """
    return GcpJobParams(**params)


def gcp_job_template(params: GcpJobParams) -> "batch_v1.Job":
    ensure_gcp_client()

    # https://github.com/GoogleCloudPlatform/python-docs-samples/blob/main/batch/create/create_with_ssd.py
    task = batch_v1.TaskSpec()
    task.max_retry_count = params.retry_count
    task.max_run_duration = f"{params.walltime_limit}s"

    ssd_name = params.ssd_name or "pulsar_staging"

    volume = batch_v1.Volume()
    volume.device_name = ssd_name
    mount_path = f"/mnt/disks/{ssd_name}"
    volume.mount_path = mount_path
    task.volumes = [volume]

    # override the staging directory since we cannot set the location of this mount path
    # the way we can in K8S based on @jmchilton's initial testing.
    environment = batch_v1.Environment(
        variables={
            "PULSAR_CONFIG_OVERRIDE_STAGING_DIRECTORY": mount_path,
        }
    )
    task.environment = environment

    # Tasks are grouped inside a job using TaskGroups.
    # Currently, it's possible to have only one task group.
    group = batch_v1.TaskGroup()
    group.task_count = 1
    group.task_spec = task

    disk = batch_v1.AllocationPolicy.Disk()
    disk.type_ = "local-ssd"
    # The size of all the local SSDs in GB. Each local SSD is 375 GB,
    # so this value must be a multiple of 375 GB.
    # For example, for 2 local SSDs, set this value to 750 GB.
    disk.size_gb = params.disk_size
    assert disk.size_gb % 375 == 0

    # Policies are used to define on what kind of virtual machines the tasks will run on.
    # The allowed number of local SSDs depends on the machine type for your job's VMs.
    # In this case, we tell the system to use "n1-standard-1" machine type, which require to attach local ssd manually.
    # Read more about local disks here: https://cloud.google.com/compute/docs/disks/local-ssd#lssd_disk_options
    policy = batch_v1.AllocationPolicy.InstancePolicy()
    policy.machine_type = params.machine_type

    attached_disk = batch_v1.AllocationPolicy.AttachedDisk()
    attached_disk.new_disk = disk
    attached_disk.device_name = ssd_name
    policy.disks = [attached_disk]

    instances = batch_v1.AllocationPolicy.InstancePolicyOrTemplate()
    instances.policy = policy

    allocation_policy = batch_v1.AllocationPolicy()
    allocation_policy.instances = [instances]

    job = batch_v1.Job()
    job.task_groups = [group]
    job.allocation_policy = allocation_policy
    job.labels = params.labels or {}
    # We use Cloud Logging as it's an out of the box available option
    job.logs_policy = batch_v1.LogsPolicy()
    job.logs_policy.destination = batch_v1.LogsPolicy.Destination.CLOUD_LOGGING  # type: ignore[assignment]

    return job


def gcp_job_request(params: GcpJobParams, job: "batch_v1.Job", job_name: str) -> "batch_v1.CreateJobRequest":
    create_request = batch_v1.CreateJobRequest()
    create_request.job = job
    create_request.job_id = job_name
    region = params.region
    project_id = params.project_id
    create_request.parent = f"projects/{project_id}/locations/{region}"
    return create_request


def container_command_to_gcp_runnable(name: str, container: CoexecutionContainerCommand) -> "batch_v1.Runnable":
    runnable = batch_v1.Runnable()
    runnable.container = batch_v1.Runnable.Container()
    runnable.container.image_uri = container.image
    runnable.container.commands = [container.command] + container.args
    # ports not supported currently
    return runnable


def gcp_galaxy_instance_id(destination_params: Dict[str, str]) -> Optional[str]:
    return destination_params.get("galaxy_instance_id")


class BasicAuth(BaseModel):
    username: str = Field(..., description="Username for basic authentication.")
    password: str = Field(..., description="Password for basic authentication.")


class TesJobParams(TesResources):
    tes_url: str = Field(..., description="URL of the TES service.")
    authorization: Literal["none", "basic"] = Field(
        "none", description="Authorization type for TES service."
    )
    basic_auth: Optional[BasicAuth] = Field(None, description="Authorization for TES service.")


def parse_tes_job_params(params: dict) -> TesJobParams:
    """
    Parse GCP job parameters parameters from a dictionary (e.g., Galaxy's job destination/environment params).
    """
    legacy_style_keys = {
        "tes_cpu_cores": "cpu_cores",
        "tes_preemptible": "preemptible",
        "tes_ram_gb": "ram_gb",
        "tes_disk_gb": "disk_gb",
        "tes_zones": "zones",
        "tes_backend_parameters": "backend_parameters",
        "tes_backend_parameters_strict": "backend_parameters_strict",
        "tes_galaxy_instance_id": "galaxy_instance_id",
    }
    expanded_params = {}
    for key, value in params.items():
        if key in legacy_style_keys:
            new_key = legacy_style_keys[key]
            expanded_params[new_key] = value
        else:
            expanded_params[key] = value

    if "zones" in expanded_params:
        expanded_params["zones"] = listify(expanded_params["zones"])

    return TesJobParams(**expanded_params)


def tes_client_from_params(tes_params: TesJobParams) -> TesClient:
    tes_url = tes_params.tes_url
    assert tes_url
    auth_type = tes_params.authorization  # Default to "none"

    headers = {}

    if auth_type == "basic":
        basic_auth = tes_params.basic_auth
        username = basic_auth.username if basic_auth else None
        password = basic_auth.password if basic_auth else None
        if username and password:
            auth_string = f"{username}:{password}"
            auth_base64 = base64.b64encode(auth_string.encode()).decode()
            headers["Authorization"] = f"Basic {auth_base64}"

    return TesClient(url=tes_url, headers=headers)


def tes_resources(tes_params: TesJobParams) -> TesResources:
    # TesJobParams subclasses it so just pass through as is.
    return tes_params



import shutil

import pkg_resources


def main():
    config_path = pkg_resources.resource_filename('pulsar.scripts', 'cvmfs_data/default.local')
    key_path = pkg_resources.resource_filename('pulsar.scripts', 'cvmfs_data/data.galaxyproject.org.pub')
    domain_path = pkg_resources.resource_filename('pulsar.scripts', 'cvmfs_data/galaxyproject.org.conf')

    # copy2?
    shutil.copy(config_path, "/etc/cvmfs/")
    shutil.copy(key_path, "/etc/cvmfs/keys/")
    shutil.copy(domain_path, "/etc/cvmfs/domain.d/")



"""
This module and its submodules contains utilities for running external
processes and interfacing with job managers. This module should contain
functionality shared between Galaxy and the Pulsar.
"""
from galaxy.util.bunch import Bunch

from .kill import kill_pid

runner_states = Bunch(
    WALLTIME_REACHED="walltime_reached",
    MEMORY_LIMIT_REACHED="memory_limit_reached",
    JOB_OUTPUT_NOT_RETURNED_FROM_CLUSTER="Job output not returned from cluster",
    UNKNOWN_ERROR="unknown_error",
    GLOBAL_WALLTIME_REACHED="global_walltime_reached",
    OUTPUT_SIZE_LIMIT="output_size_limit",
    TOOL_DETECT_ERROR="tool_detected",  # job runner interaction worked fine but the tool indicated error
)


__all__ = (
    "kill_pid",
    "runner_states",
)



import contextlib
import logging
import threading

try:
    from drmaa import (
        JobControlAction,
        Session,
    )
except OSError as e:
    LOAD_ERROR_MESSAGE = "OSError - problem loading shared library [%s]." % e
    Session = None
except ImportError as e:
    LOAD_ERROR_MESSAGE = "ImportError - problem importing library (`pip install drmaa` may fix this) [%s]." % e
    # Will not be able to use DRMAA
    Session = None
except RuntimeError as e:
    LOAD_ERROR_MESSAGE = "RuntimeError - problem importing library (setting DRMAA_LIBRARY_PATH may fix this) [%s]." % e
    Session = None

NO_DRMAA_MESSAGE = "Attempt to use DRMAA, but DRMAA Python library cannot be loaded. "

log = logging.getLogger(__name__)


class DrmaaSessionFactory:
    """
    Abstraction used to production DrmaaSession wrappers.
    """
    def __init__(self):
        self.session_constructor = Session

    def get(self, **kwds):
        session_constructor = self.session_constructor
        if session_constructor is None:
            raise Exception(NO_DRMAA_MESSAGE + LOAD_ERROR_MESSAGE)
        return DrmaaSession(session_constructor, **kwds)


class DrmaaSession:
    """
    Abstraction around `drmaa` module `Session` objects.
    """
    session_lock = threading.Lock()
    session_count = 0
    session = None

    def __init__(self, session_constructor, **kwds):
        with self._session_lock():
            if DrmaaSession.session is None:
                if DrmaaSession.session_count != 0:
                    log.warn("DrmaaSession.session is None but session_count is non-zero - logic error occurred.")
                log.debug("Initializing DRMAA session from thread %s", threading.current_thread().name)
                DrmaaSession.session = session_constructor()
                DrmaaSession.session.initialize()
            DrmaaSession.session_count += 1

    @contextlib.contextmanager
    def _session_lock(self):
        with DrmaaSession.session_lock:
            yield

    def run_job(self, **kwds):
        """
        Create a DRMAA job template, populate with specified properties,
        run the job, and return the external_job_id.
        """
        template = DrmaaSession.session.createJobTemplate()
        try:
            for key in kwds:
                setattr(template, key, kwds[key])
            with DrmaaSession.session_lock:
                return DrmaaSession.session.runJob(template)
        finally:
            DrmaaSession.session.deleteJobTemplate(template)

    def kill(self, external_job_id):
        with DrmaaSession.session_lock:
            return DrmaaSession.session.control(str(external_job_id), JobControlAction.TERMINATE)

    def job_status(self, external_job_id):
        return DrmaaSession.session.jobStatus(str(external_job_id))

    def close(self):
        with self._session_lock():
            if DrmaaSession.session_count == 0:
                log.warn("close() called with zero active session counted - logic error.")
                return

            DrmaaSession.session_count -= 1
            if DrmaaSession.session_count == 0:
                if DrmaaSession.session is None:
                    log.warn("close() called with a non-zero session count but no session is defined.")
                    return

                DrmaaSession.session.exit()
                DrmaaSession.session = None


__all__ = ['DrmaaSessionFactory']



"""Small utility for bootstrapping a Conda environment for Pulsar.

This should probably be moved into galaxy-tool-util.
"""

import os.path
import sys
from argparse import ArgumentParser

from galaxy.tool_util.deps.conda_util import (
    CondaContext,
    install_conda,
)
from galaxy.util import safe_makedirs


def main(argv=None):
    mod_docstring = sys.modules[__name__].__doc__
    arg_parser = ArgumentParser(description=mod_docstring)
    arg_parser.add_argument("--conda_prefix", required=True)
    args = arg_parser.parse_args(argv)
    conda_prefix = args.conda_prefix
    safe_makedirs(os.path.dirname(conda_prefix))
    conda_context = CondaContext(
        conda_prefix=conda_prefix,
    )
    install_conda(conda_context)


if __name__ == "__main__":
    main()



""" This module and submodules contain code for interfacing the Apache Mesos framework.
"""
try:
    from mesos.interface import (
        Executor,
        Scheduler,
    )
    from mesos.native import (
        MesosExecutorDriver,
        MesosSchedulerDriver,
    )
except ImportError:
    Scheduler = object
    Executor = object
    MesosSchedulerDriver = None
    MesosExecutorDriver = None
try:
    from mesos.interface import mesos_pb2
except ImportError:
    mesos_pb2 = None

NO_MESOS_EXCEPTION = "Failed to import mesos module, please install mesos properly."
NO_MESOS_PROTO_EXCEPTION = "Failed to import mesos_pbs module, please install mesos properly."


def ensure_mesos_libs():
    """ Raise import error if mesos is not actually available. Original
    import errors above supressed because mesos is meant as an optional
    dependency for the Pulsar.
    """

    if MesosSchedulerDriver is None:
        raise Exception(NO_MESOS_EXCEPTION)

    if mesos_pb2 is None:
        raise Exception(NO_MESOS_PROTO_EXCEPTION)



"""State management for pulsar-relay message consumers.

Similar to QueueState for AMQP, this manages the lifecycle of relay
consumer threads on the Pulsar server side.
"""


class RelayState:
    """Manages state for pulsar-relay message consumers.

    This object is passed to consumer loops and used to signal when
    they should stop processing messages.
    """

    def __init__(self):
        """Initialize relay state."""
        self.active = True
        self.threads = []

    def deactivate(self):
        """Mark the relay state as inactive, signaling consumers to stop."""
        self.active = False

    def join(self, timeout=None):
        """Join all consumer threads.

        Args:
            timeout: Optional timeout in seconds for joining threads
        """
        import logging
        log = logging.getLogger(__name__)

        for t in self.threads:
            t.join(timeout)
            if t.is_alive():
                log.warning("Failed to join relay consumer thread [%s].", t)



import functools
import logging
import threading

from galaxy.util import (
    asbool,
    mask_password_from_url,
)

from pulsar import manager_endpoint_util
from pulsar.client import amqp_exchange_factory

log = logging.getLogger(__name__)


TYPED_PARAMS = {
    "amqp_consumer_timeout": lambda val: None if str(val) == "None" else float(val),
    "amqp_publish_timeout": lambda val: None if str(val) == "None" else float(val),
    "amqp_publish_retry": asbool,
    "amqp_publish_retry_max_retries": int,
    "amqp_publish_retry_interval_start": int,
    "amqp_publish_retry_interval_step": int,
    "amqp_publish_retry_interval_max": int,
}


def get_exchange(connection_string, manager_name, conf):
    # HACK: Fixup non-string parameters - ultimately this should reuse spec
    # stuff from Galaxy.
    for param, to_type in TYPED_PARAMS.items():
        if param in conf:
            val = conf[param]
            conf[param] = to_type(val)

    pulsar_exchange = amqp_exchange_factory.get_exchange(
        connection_string,
        manager_name,
        conf
    )
    return pulsar_exchange


def bind_manager_to_queue(manager, queue_state, connection_string, conf):
    manager_name = manager.name
    log.info("bind_manager_to_queue called for [{}] and manager [{}]".format(mask_password_from_url(connection_string), manager_name))
    pulsar_exchange = get_exchange(connection_string, manager_name, conf)

    process_setup_messages = functools.partial(__process_setup_message, manager)
    process_kill_messages = functools.partial(__process_kill_message, manager)
    process_status_messages = functools.partial(__process_status_message, manager)

    def drain(callback, name):
        __drain(name, queue_state, pulsar_exchange, callback)
        log.info("Finished consuming %s queue - no more messages will be processed." % (name))

    if conf.get("message_queue_consume", True):
        setup_thread = start_setup_consumer(pulsar_exchange, functools.partial(drain, process_setup_messages, "setup"))
        kill_thread = start_kill_consumer(pulsar_exchange, functools.partial(drain, process_kill_messages, "kill"))
        status_thread = start_status_consumer(pulsar_exchange, functools.partial(drain, process_status_messages, "status"))
        if hasattr(queue_state, "threads"):
            queue_state.threads.extend([setup_thread, kill_thread, status_thread])
        if conf.get("amqp_acknowledge", False):
            status_update_ack_thread = start_status_update_ack_consumer(pulsar_exchange, functools.partial(drain, None, "status_update_ack"))
            getattr(queue_state, 'threads', []).append(status_update_ack_thread)

    # TODO: Think through job recovery, jobs shouldn't complete until after bind
    # has occurred.
    def bind_on_status_change(new_status, job_id):
        job_id = job_id or 'unknown'
        try:
            message = "Publishing Pulsar state change with status {} for job_id {}".format(new_status, job_id)
            log.debug(message)
            payload = manager_endpoint_util.full_status(manager, new_status, job_id)
            pulsar_exchange.publish("status_update", payload)
        except Exception:
            log.exception("Failure to publish Pulsar state change for job_id %s." % job_id)
            raise

    if conf.get("message_queue_publish", True):
        manager.set_state_change_callback(bind_on_status_change)


def __start_consumer(name, exchange, target):
    exchange_url = mask_password_from_url(exchange.url)
    thread_name = "consume-{}-{}".format(name, exchange_url)
    thread = threading.Thread(name=thread_name, target=target)
    # TODO: If the shutdown code is actually called make this
    # not a daemon.
    thread.daemon = True
    thread.start()
    return thread


start_setup_consumer = functools.partial(__start_consumer, "setup")
start_kill_consumer = functools.partial(__start_consumer, "kill")
start_status_consumer = functools.partial(__start_consumer, "status")
start_status_update_ack_consumer = functools.partial(__start_consumer, "status_update_ack")


def __drain(name, queue_state, pulsar_exchange, callback):
    pulsar_exchange.consume(name, callback=callback, check=queue_state)


def __processes_message(f):

    @functools.wraps(f)
    def process_message(manager, body, message):
        if message.acknowledged:
            log.info("Message is already acknowledged (by an upstream callback?), Pulsar will not handle this message")
            return

        job_id = None
        try:
            job_id = __client_job_id_from_body(body)
            assert job_id, 'Could not parse job id from body: %s' % body
            f(manager, body, job_id)
        except Exception:
            job_id = job_id or 'unknown'
            log.exception("Failed to process message with function {} for job_id {}".format(f.__name__, job_id))
        message.ack()

    return process_message


@__processes_message
def __process_kill_message(manager, body, job_id):
    manager.kill(job_id)


@__processes_message
def __process_setup_message(manager, body, job_id):
    manager_endpoint_util.submit_job(manager, body)


@__processes_message
def __process_status_message(manager, body, job_id):
    manager.trigger_state_change_callback(job_id)


def __client_job_id_from_body(body):
    job_id = body.get("job_id", None)
    return job_id



import shelve
import traceback

from threading import Lock


class PersistenceStore:

    def __init__(self, filename, require_sync=True):
        self.shelf_filename = filename
        self.__require_sync = require_sync
        self.__open_shelf()
        self.__shelf_lock = Lock()

    def __open_shelf(self):
        self.shelf = shelve.open(self.shelf_filename, writeback=self.__require_sync) if self.shelf_filename else None

    def close(self):
        self.shelf.close()

    def _sync_if_needed(self):
        if self.__require_sync:
            self.shelf.sync()

    def _lock(self):
        return self.__shelf_lock

    def _with_lock(self, func, suppress_exception=True):
        if self.shelf is not None:
            with self._lock():
                try:
                    return func()
                except Exception:
                    traceback.print_exc()
                    if not suppress_exception:
                        raise



"""
This file is also being used by the GalaxyCloudRunner (gcr) Docker image.
"""

from getpass import getuser
from multiprocessing import cpu_count
from socket import gethostname
from string import Template

SLURM_CONFIG_TEMPLATE = '''
# slurm.conf file generated by configurator.html.
# Put this file on all nodes of your cluster.
# See the slurm.conf man page for more information.
#
ControlMachine=$hostname
#ControlAddr=
#BackupController=
#BackupAddr=
#
AuthType=auth/munge
CacheGroups=0
#CheckpointType=checkpoint/none
CryptoType=crypto/munge
MpiDefault=none
#PluginDir=
#PlugStackConfig=
#PrivateData=jobs
ProctrackType=proctrack/pgid
#Prolog=
#PrologSlurmctld=
#PropagatePrioProcess=0
#PropagateResourceLimits=
#PropagateResourceLimitsExcept=
ReturnToService=1
#SallocDefaultCommand=
SlurmctldPidFile=/var/run/slurmctld.pid
SlurmctldPort=6817
SlurmdPidFile=/var/run/slurmd.pid
SlurmdPort=6818
SlurmdSpoolDir=/tmp/slurmd
SlurmUser=$user
#SlurmdUser=root
#SrunEpilog=
#SrunProlog=
StateSaveLocation=/tmp
SwitchType=switch/none
#TaskEpilog=
TaskPlugin=task/none
#TaskPluginParam=
#TaskProlog=
InactiveLimit=0
KillWait=30
MinJobAge=300
#OverTimeLimit=0
SlurmctldTimeout=120
SlurmdTimeout=300
#UnkillableStepTimeout=60
#VSizeFactor=0
Waittime=0
FastSchedule=1
SchedulerType=sched/backfill
SchedulerPort=7321
SelectType=select/linear
#SelectTypeParameters=
AccountingStorageType=accounting_storage/none
#AccountingStorageUser=
AccountingStoreJobComment=YES
ClusterName=cluster
#DebugFlags=
#JobCompHost=
#JobCompLoc=
#JobCompPass=
#JobCompPort=
JobCompType=jobcomp/none
#JobCompUser=
JobAcctGatherFrequency=30
JobAcctGatherType=jobacct_gather/none
SlurmctldDebug=3
#SlurmctldLogFile=
SlurmdDebug=3
#SlurmdLogFile=
NodeName=$hostname CPUs=$cpus State=UNKNOWN
PartitionName=debug Nodes=$hostname Default=YES MaxTime=INFINITE State=UP
'''


def main():
    template_params = {"hostname": gethostname(),
                       "user": getuser(),
                       "cpus": cpu_count()}
    config_contents = Template(SLURM_CONFIG_TEMPLATE).substitute(template_params)
    open("/etc/slurm-llnl/slurm.conf", "w").write(config_contents)


if __name__ == "__main__":
    main()



import re
from enum import Enum
from os import sep
from os.path import (
    basename,
    dirname,
    exists,
    join,
)
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Union,
)

from galaxy.tool_util.parser.output_collection_def import dataset_collection_description
from galaxy.util.bunch import Bunch

from ..util import PathHelper

if TYPE_CHECKING:
    from galaxy.tool_util.parser.interface import RequiredFiles

COMMAND_VERSION_FILENAME = "COMMAND_VERSION"
DEFAULT_DYNAMIC_COLLECTION_PATTERN = [
    "|".join(
        [
            r"primary_.*",
            r"galaxy.json",
            r"metadata_.*",
            r"dataset_\d+\.dat",
            r"__instrument_.*",
            r"dataset_\d+_files.+",
            r"outputs_populated/.*",
            r"tool_stdout",
            r"tool_stderr",
        ]
    )
]
EXTENDED_METADATA_DYNAMIC_COLLECTION_PATTERN = [
    "|".join(
        [
            r"outputs_populated/.*",
        ]
    )
]


class DynamicFileSourceType(str, Enum):
    legacy_galaxy = "legacy_galaxy"  # older stlye galaxy.json - line per entry
    galaxy = "galaxy"  # modern galaxy.json - single json
    # TODO: cwl outputs file, etc...


class ClientJobDescription:
    """ A description of how client views job - command_line, inputs, etc..

    **Parameters**

    command_line : str
        The local command line to execute, this will be rewritten for
        the remote server.
    tool :
        The Galaxy tool to execute.
    config_files : list
        List of Galaxy 'configfile's produced for this job. These will
        be rewritten and sent to remote server.
    input_files :  list
        List of input files used by job. These will be transferred and
        references rewritten.
    client_outputs : ClientOutputs
        Description of outputs produced by job (at least output files along
        with optional version string and working directory outputs.
    working_directory : str
        Local path created by Galaxy for running this job (job_wrapper.tool_working_directory).
    metadata_directory : str
        Local path created by Galaxy for running this job (job_wrapper.working_directory).
    dependencies_description : list
        galaxy.tools.deps.dependencies.DependencyDescription object describing
        tool dependency context for remote depenency resolution.
    env: list
        List of dict object describing environment variables to populate.
    arbitrary_files : dict()
        Additional non-input, non-tool, non-config, non-working directory files
        to transfer before staging job. This is most likely data indices but
        can be anything. For now these are copied into staging working
        directory but this will be reworked to find a better, more robust
        location.
    rewrite_paths : boolean
        Indicates whether paths should be rewritten in job inputs (command_line
        and config files) while staging files).
    """

    def __init__(
        self,
        command_line,
        tool=None,
        config_files=None,
        input_files=None,
        client_inputs=None,
        client_outputs=None,
        working_directory=None,
        metadata_directory=None,
        dependencies_description=None,
        env=[],
        arbitrary_files=None,
        job_directory_files=None,
        tool_directory_required_files: Union["RequiredFiles", None] = None,
        rewrite_paths=True,
        touch_outputs=None,
        container=None,
        remote_pulsar_app_config=None,
        guest_ports=None,
    ):
        self.tool = tool
        self.command_line = command_line
        self.config_files = config_files or []
        if input_files is not None:
            # Deprecated input but provided for backward compatibility.
            assert client_inputs is None
            client_inputs = ClientInputs.for_simple_input_paths(input_files)
        self.client_inputs = client_inputs or ClientInputs([])
        self.client_outputs = client_outputs or ClientOutputs()
        self.working_directory = working_directory
        self.metadata_directory = metadata_directory
        self.dependencies_description = dependencies_description
        self.env = env
        self.job_directory_files = job_directory_files or []
        self.tool_directory_required_files = tool_directory_required_files
        self.rewrite_paths = rewrite_paths
        self.arbitrary_files = arbitrary_files or {}
        self.touch_outputs = touch_outputs or []
        self.container = container
        self.guest_ports = guest_ports
        self.remote_pulsar_app_config = remote_pulsar_app_config

    @property
    def input_files(self):
        # Deprecated but provided for backward compatibility.
        input_files = []
        for client_input in self.client_inputs:
            if client_input.input_type == CLIENT_INPUT_PATH_TYPES.INPUT_PATH:
                input_files.append(client_input.path)
        return input_files

    @property
    def output_files(self):
        return self.client_outputs.output_files

    @property
    def version_file(self):
        return self.client_outputs.version_file

    @property
    def tool_dependencies(self):
        if not self.remote_dependency_resolution:
            return None
        return dict(
            requirements=(self.tool.requirements or []),
            installed_tool_dependencies=(self.tool.installed_tool_dependencies or [])
        )


class ClientInputs:
    """Abstraction describing input datasets for a job."""

    def __init__(self, client_inputs):
        self.client_inputs = client_inputs

    def __iter__(self):
        return iter(self.client_inputs)

    @staticmethod
    def for_simple_input_paths(input_files):
        # Legacy: just assume extra files path based on inputs, probably not
        # the best behavior - ignores object store for instance.
        client_inputs = []
        for input_file in input_files:
            client_inputs.append(ClientInput(input_file, CLIENT_INPUT_PATH_TYPES.INPUT_PATH))
            files_path = "%s_files" % input_file[0:-len(".dat")]
            if exists(files_path):
                client_inputs.append(ClientInput(files_path, CLIENT_INPUT_PATH_TYPES.INPUT_EXTRA_FILES_PATH))

        return ClientInputs(client_inputs)


CLIENT_INPUT_PATH_TYPES = Bunch(
    INPUT_PATH="input_path",
    INPUT_EXTRA_FILES_PATH="input_extra_files_path",
    INPUT_METADATA_PATH="input_metadata_path",
)


class ClientInput:

    def __init__(self, path, input_type, object_store_ref=None):
        self.path = path
        self.input_type = input_type
        self.object_store_ref = object_store_ref

    @property
    def action_source(self):
        return {"path": self.path, "object_store_ref": self.object_store_ref}


class ClientOutputs:
    """ Abstraction describing the output datasets EXPECTED by the Galaxy job
    runner client.
    """

    def __init__(
        self,
        working_directory=None,
        output_files=[],
        work_dir_outputs=None,
        version_file=None,
        dynamic_outputs=None,
        metadata_directory=None,
        job_directory=None,
        dynamic_file_sources=None,
        dataset_collector_descriptions=None,
    ):
        self.working_directory = working_directory
        self.metadata_directory = metadata_directory
        self.work_dir_outputs = work_dir_outputs or []
        self.output_files = output_files or []
        self.version_file = version_file
        self.dynamic_outputs = dynamic_outputs or DEFAULT_DYNAMIC_COLLECTION_PATTERN
        self.job_directory = job_directory
        self.dynamic_file_sources = dynamic_file_sources
        self.dataset_collector_descriptions = dataset_collector_descriptions or []
        self.__dynamic_patterns = list(map(re.compile, self.dynamic_outputs))

    def to_dict(self):
        return dict(
            working_directory=self.working_directory,
            metadata_directory=self.metadata_directory,
            job_directory=self.job_directory,
            work_dir_outputs=self.work_dir_outputs,
            output_files=self.output_files,
            version_file=self.version_file,
            dynamic_outputs=self.dynamic_outputs,
            dynamic_file_sources=self.dynamic_file_sources,
            dataset_collector_descriptions=self.dataset_collector_descriptions,
        )

    @staticmethod
    def from_dict(config_dict):
        return ClientOutputs(
            working_directory=config_dict.get('working_directory'),
            metadata_directory=config_dict.get('metadata_directory'),
            work_dir_outputs=config_dict.get('work_dir_outputs'),
            output_files=config_dict.get('output_files'),
            version_file=config_dict.get('version_file'),
            dynamic_outputs=config_dict.get('dynamic_outputs'),
            dynamic_file_sources=config_dict.get('dynamic_file_sources'),
            job_directory=config_dict.get('job_directory'),
            dataset_collector_descriptions=config_dict.get('dataset_collector_descriptions'),
        )

    def dynamic_match(self, filename):

        # First try matching with dataset collectors if available (supports directory, recurse, match_relative_path)
        if self.dataset_collector_descriptions:
            filename_path = Path(filename)

            for desc_dict in self.dataset_collector_descriptions:
                desc = dataset_collection_description(**desc_dict)
                directory = desc.directory

                if directory:
                    directory_path = Path(directory)

                    # Check if filename is within the specified directory
                    try:
                        relative_path = filename_path.resolve().relative_to(directory_path.resolve())
                    except ValueError:
                        # filename is not within directory, continue to next collector
                        continue

                    # Determine what to match against the pattern
                    if desc.match_relative_path:
                        # For match_relative_path, include parent directories in the match
                        match_string = str(relative_path)
                    else:
                        # Otherwise, just match the final component (filename)
                        match_string = relative_path.name if len(relative_path.parts) == 1 else str(relative_path)

                    # Match against the pattern
                    if re.match(desc.pattern, match_string):
                        return True
                else:
                    # No directory restriction, match directly against filename
                    if re.match(desc.pattern, filename):
                        return True

        # Collect DEFAULT_DYNAMIC_COLLECTION_PATTERN and provided_metadata_file patterns
        for pattern in self.__dynamic_patterns:
            if pattern.match(filename):
                return True
        return False


class PulsarOutputs:
    """ Abstraction describing the output files PRODUCED by the remote Pulsar
    server. """

    def __init__(
        self,
        working_directory_contents,
        output_directory_contents,
        metadata_directory_contents,
        job_directory_contents,
        remote_separator=sep,
        realized_dynamic_file_sources=None,  # list of dicts with keys - name, path, contents. (client def + realized contents)
    ):
        self.working_directory_contents = working_directory_contents
        self.output_directory_contents = output_directory_contents
        self.metadata_directory_contents = metadata_directory_contents
        self.job_directory_contents = job_directory_contents
        self.realized_dynamic_file_sources = realized_dynamic_file_sources
        self.path_helper = PathHelper(remote_separator)

    @staticmethod
    def from_status_response(complete_response):
        # Default to None instead of [] to distinguish between empty contents and it not set
        # by the Pulsar - older Pulsar instances will not set these in complete response.
        working_directory_contents = complete_response.get("working_directory_contents")
        output_directory_contents = complete_response.get("outputs_directory_contents")
        metadata_directory_contents = complete_response.get("metadata_directory_contents")
        job_directory_contents = complete_response.get("job_directory_contents")
        # Older (pre-2014) Pulsar servers will not include separator in response,
        # so this should only be used when reasoning about outputs in
        # subdirectories (which was not previously supported prior to that).
        remote_separator = complete_response.get("system_properties", {}).get("separator", sep)
        realized_dynamic_file_sources = complete_response.get("realized_dynamic_file_sources", None)
        return PulsarOutputs(
            working_directory_contents,
            output_directory_contents,
            metadata_directory_contents,
            job_directory_contents,
            remote_separator,
            realized_dynamic_file_sources=realized_dynamic_file_sources,
        )

    def has_output_file(self, output_file):
        return basename(output_file) in self.output_directory_contents

    def output_extras(self, output_file):
        """
        Returns dict mapping local path to remote name.
        """
        output_directory = dirname(output_file)

        def local_path(name):
            return join(output_directory, self.path_helper.local_name(name))

        files_directory = "{}_files{}".format(basename(output_file)[0:-len(".dat")], self.path_helper.separator)
        names = filter(lambda o: o.startswith(files_directory), self.output_directory_contents)
        return dict(map(lambda name: (local_path(name), name), names))



import os
from logging import getLogger
from subprocess import (
    PIPE,
    Popen,
)
from tempfile import TemporaryFile
from time import sleep

from galaxy.util.bunch import Bunch

from . import BaseShellExec
from ....util.process_groups import (
    check_pg,
    kill_pg,
)

log = getLogger(__name__)

TIMEOUT_ERROR_MESSAGE = "Execution timed out"
TIMEOUT_RETURN_CODE = -1
DEFAULT_TIMEOUT = 60
DEFAULT_TIMEOUT_CHECK_INTERVAL = 3


class LocalShell(BaseShellExec):
    """

    >>> shell = LocalShell()
    >>> def exec_python(script, **kwds): return shell.execute(['python', '-c', script], **kwds)
    >>> exec_result = exec_python("from __future__ import print_function; print('Hello World')")
    >>> exec_result.stderr == u''
    True
    >>> exec_result.stdout.strip() == u'Hello World'
    True
    >>> exec_result.returncode
    0
    >>> exec_result = exec_python("import time; time.sleep(10)", timeout=1, timeout_check_interval=.1)
    >>> exec_result.stdout == u''
    True
    >>> exec_result.stderr == 'Execution timed out'
    True
    >>> exec_result.returncode == TIMEOUT_RETURN_CODE
    True
    >>> shell.execute('echo hi').stdout == "hi\\n"
    True
    """

    def __init__(self, **kwds):
        pass

    def execute(
        self, cmd, persist=False, timeout=DEFAULT_TIMEOUT, timeout_check_interval=DEFAULT_TIMEOUT_CHECK_INTERVAL, **kwds
    ):
        is_cmd_string = isinstance(cmd, str)
        outf = TemporaryFile()
        p = Popen(cmd, stdin=None, stdout=outf, stderr=PIPE, shell=is_cmd_string, preexec_fn=os.setpgrp)
        # check process group until timeout

        for _ in range(int(timeout / timeout_check_interval)):
            sleep(0.1)  # For fast returning commands
            if not check_pg(p.pid):
                break
            sleep(timeout_check_interval)
        else:
            kill_pg(p.pid)
            return Bunch(stdout="", stderr=TIMEOUT_ERROR_MESSAGE, returncode=TIMEOUT_RETURN_CODE)
        outf.seek(0)
        # Need to poll once to establish return code
        p.poll()
        return Bunch(stdout=_read_str(outf), stderr=_read_str(p.stderr), returncode=p.returncode)


def _read_str(stream):
    contents = stream.read()
    return contents.decode("UTF-8") if isinstance(contents, bytes) else contents


__all__ = ("LocalShell",)



# TODO: Make objects.

# Job is staging about will be queued shortly.
PREPROCESSING = "preprocessing"
# Job manager has queued this job for execution.
QUEUED = "queued"
# Job manager believes the job is currently running.
RUNNING = "running"
# Job manager has finished and postprocessing ran successfully.
COMPLETE = "complete"
# Job was cancelled
CANCELLED = "cancelled"
# Problem submitting the job, interfacing with the job manager,
# or postprocessing the job.
FAILED = "failed"
# DRM marked job as complete and job is being unstaged.
POSTPROCESSING = "postprocessing"
# Pulsar believed this job to be active but the job manager
# cannot determine a state for it.
LOST = "lost"


def is_job_done(status):
    """ Does the supplied status correspond to a finished
    job (done processing).
    """
    return status in [COMPLETE, CANCELLED, FAILED, LOST]



from pulsar.user_auth.methods.interface import AuthMethod


class AlwaysAllowAuthMethod(AuthMethod):
    """
    Always allow
    """

    def __init__(self, _config):
        pass

    auth_type = "allow_all"

    def authorize(self, authentication_info):
        return True

    def authenticate(self, job_directory):
        return {"username": "anonymous"}



"""Interface layer for Boto's batch library shared between Galaxy and Pulsar."""

BOTO3_IMPORT_MSG = (
    "The Python 'boto3' package is required to use "
    "this feature, please install it or correct the "
    "following error:\nImportError {msg!s}"
)

try:
    import boto3

except ImportError as e:
    boto3 = None
    BOTO3_IMPORT_MSG.format(msg=str(e))


def ensure_client_available() -> None:
    if boto3 is None:
        raise Exception(BOTO3_IMPORT_MSG)


def get_client(access_key_id=None, secret_access_key=None, region_name=None):
    session = boto3.Session(
        aws_access_key_id=access_key_id,
        aws_secret_access_key=secret_access_key,
        region_name=region_name,
    )
    return session.client("batch")


__all__ = (
    "ensure_client_available",
    "get_client",
)



""" This module contains the server-side only code for interfacing with
message queues. Code shared between client and server can be found in
submodules of ``pulsar.client``.
"""

import logging

from ..messaging import (
    bind_amqp,
    bind_relay,
)
from .relay_state import RelayState

log = logging.getLogger(__name__)


def bind_app(app, queue_id, conf=None):
    connection_string = __id_to_connection_string(app, queue_id)

    # Check if this is a relay connection
    if connection_string and connection_string.startswith('http://') or connection_string.startswith('https://'):
        relay_url = connection_string
        log.info("Detected relay connection string, binding to pulsar-relay at %s", relay_url)

        relay_state = RelayState()
        for manager in app.managers.values():
            bind_relay.bind_manager_to_relay(manager, relay_state, relay_url, conf or {})
        return relay_state
    else:
        # Use AMQP binding
        queue_state = QueueState()
        for manager in app.managers.values():
            bind_amqp.bind_manager_to_queue(manager, queue_state, connection_string, conf)
        return queue_state


class QueueState:
    """ Passed through to event loops, should be "non-zero" while queues should
    be active.
    """
    def __init__(self):
        self.active = True
        self.threads = []

    def deactivate(self):
        self.active = False

    def __nonzero__(self):
        return self.active

    __bool__ = __nonzero__  # Both needed Py2 v 3

    def join(self, timeout=None):
        for t in self.threads:
            t.join(timeout)
            if t.is_alive():
                log.warning("Failed to join thread [%s]." % t)


def __id_to_connection_string(app, queue_id):
    return queue_id



__version__ = '0.15.13'

PROJECT_NAME = "pulsar"
PROJECT_OWNER = PROJECT_USERAME = "galaxyproject"
PROJECT_AUTHOR = 'Galaxy Project and Community'
PROJECT_EMAIL = 'jmchilton@gmail.com'

PROJECT_URL = "https://github.com/{}/{}".format(PROJECT_OWNER, PROJECT_NAME)
RAW_CONTENT_URL = "https://raw.github.com/{}/{}/master/".format(
    PROJECT_USERAME, PROJECT_NAME
)



import logging
import time

import paramiko
from galaxy.util import (
    smart_str,
    string_as_bool,
    unicodify,
)
from galaxy.util.bunch import Bunch

from pulsar.managers.util.retry import RetryActionExecutor
from .local import LocalShell

log = logging.getLogger(__name__)
logging.getLogger("paramiko").setLevel(logging.WARNING)  # paramiko logging is very verbose

__all__ = ("RemoteShell", "SecureShell", "GlobusSecureShell", "ParamikoShell")


class RemoteShell(LocalShell):
    def __init__(self, rsh="rsh", rcp="rcp", hostname="localhost", username=None, options=None, **kwargs):
        super().__init__(**kwargs)
        self.rsh = rsh
        self.rcp = rcp
        self.hostname = hostname
        self.username = username
        self.options = options
        self.sessions = {}

    def execute(self, cmd, persist=False, timeout=60):
        # TODO: implement persistence
        fullcmd = [self.rsh]
        if self.options:
            fullcmd.extend(self.options)
        if self.username:
            fullcmd.extend(["-l", self.username])
        fullcmd.extend([self.hostname, cmd])
        return super().execute(fullcmd, persist, timeout)


class SecureShell(RemoteShell):
    def __init__(self, rsh="ssh", rcp="scp", private_key=None, port=None, strict_host_key_checking=True, **kwargs):
        options = []
        if not string_as_bool(strict_host_key_checking):
            options.extend(["-o", "StrictHostKeyChecking=no", "-o", "UserKnownHostsFile=/dev/null"])
        options.extend(["-o", "ConnectTimeout=60"])
        if private_key:
            options.extend(["-i", private_key])
        if port:
            options.extend(["-p", str(port)])
        super().__init__(rsh=rsh, rcp=rcp, options=options, **kwargs)


class ParamikoShell:
    def __init__(
        self,
        username,
        hostname,
        password=None,
        private_key=None,
        port=22,
        timeout=60,
        strict_host_key_checking=True,
        **kwargs,
    ):
        self.username = username
        self.hostname = hostname
        self.password = password
        self.private_key = private_key
        self.port = int(port) if port else port
        self.timeout = int(timeout) if timeout else timeout
        self.strict_host_key_checking = string_as_bool(strict_host_key_checking)
        self.ssh = None
        self.retry_action_executor = RetryActionExecutor(max_retries=100, interval_max=300)
        self.connect()

    def connect(self):
        log.info("Attempting establishment of new paramiko SSH channel")
        self.ssh = paramiko.SSHClient()
        self.ssh.set_missing_host_key_policy(
            paramiko.RejectPolicy() if self.strict_host_key_checking else paramiko.WarningPolicy()
        )
        self.ssh.load_system_host_keys()
        self.ssh.connect(
            hostname=self.hostname,
            port=self.port,
            username=self.username,
            password=self.password,
            key_filename=self.private_key,
            timeout=self.timeout,
        )

    def execute(self, cmd, timeout=60):
        def retry():
            try:
                _, stdout, stderr = self._execute(cmd, timeout)
            except paramiko.SSHException as e:
                log.error(e)
                time.sleep(10)
                self.connect()
                _, stdout, stderr = self._execute(cmd, timeout)
            return stdout, stderr

        stdout, stderr = self.retry_action_executor.execute(retry)
        return_code = stdout.channel.recv_exit_status()
        return Bunch(stdout=unicodify(stdout.read()), stderr=unicodify(stderr.read()), returncode=return_code)

    def _execute(self, cmd, timeout):
        return self.ssh.exec_command(smart_str(cmd), timeout=timeout)


class GlobusSecureShell(SecureShell):
    def __init__(self, rsh="gsissh", rcp="gsiscp", **kwargs):
        super().__init__(rsh=rsh, rcp=rcp, **kwargs)



import requests
import base64
import json
import jwt
import re
from cryptography.hazmat.backends import default_backend
from cryptography.x509 import load_der_x509_certificate

from pulsar.user_auth.methods.interface import AuthMethod

import logging

log = logging.getLogger(__name__)


def get_token(job_directory, provider):
    log.debug("Getting OIDC token for provider " + provider + " from Galaxy")
    endpoint = job_directory.load_metadata("launch_config")["token_endpoint"]
    endpoint = endpoint + "&provider=" + provider
    r = requests.get(url=endpoint)
    return r.text


class OIDCAuth(AuthMethod):
    """
    Authorization based on OIDC tokens
    """
    auth_type = "oidc"

    def __init__(self, config):
        try:
            self._provider = config["oidc_provider"]
            self._jwks_url = config["oidc_jwks_url"]
            self._username_in_token = config["oidc_username_in_token"]
            self._username_template = config["oidc_username_template"]

        except Exception as e:
            raise Exception("cannot read OIDCAuth configuration") from e

    def _verify_token(self, token):
        try:
            # Obtain appropriate cert from JWK URI
            key_set = requests.get(self._jwks_url, timeout=5)
            encoded_header, rest = token.split('.', 1)
            headerobj = json.loads(base64.b64decode(encoded_header + '==').decode('utf8'))
            key_id = headerobj['kid']
            for key in key_set.json()['keys']:
                if key['kid'] == key_id:
                    x5c = key['x5c'][0]
                    break
            else:
                raise jwt.DecodeError('Cannot find kid ' + key_id)
            cert = load_der_x509_certificate(base64.b64decode(x5c), default_backend())
            # Decode token (exp date is checked automatically)
            decoded_token = jwt.decode(
                token,
                key=cert.public_key(),
                algorithms=['RS256'],
                options={'exp': True, 'verify_aud': False}
            )
            return decoded_token
        except Exception as error:
            raise Exception("Error verifying jwt token") from error

    def authorize(self, authentication_info):
        raise NotImplementedError("authorization not implemented for this class")

    def authenticate(self, job_directory):
        token = get_token(job_directory, self._provider)

        decoded_token = self._verify_token(token)
        user = decoded_token[self._username_in_token]
        user = re.match(self._username_template, user).group(0)
        return {"username": user}



"""
JWT authentication manager for pulsar-relay.

Handles token acquisition, caching, and automatic refresh.
"""
import logging
import threading
from typing import cast, Optional
from datetime import datetime, timedelta

import requests

log = logging.getLogger(__name__)


class RelayAuthManager:
    """Manages JWT authentication tokens for pulsar-relay communication.

    Features:
    - Thread-safe token caching
    - Automatic token refresh before expiry
    - Lazy authentication (only authenticates when needed)
    """

    def __init__(self, relay_url: str, username: str, password: str):
        """Initialize the authentication manager.

        Args:
            relay_url: Base URL of the pulsar-relay server
            username: Username for authentication
            password: Password for authentication
        """
        self.relay_url = relay_url.rstrip('/')
        self.username = username
        self.password = password

        self._token: Optional[str] = None
        self._token_expiry: Optional[datetime] = None
        self._lock = threading.Lock()

        # Refresh token 5 minutes before expiry
        self._refresh_buffer_seconds = 300

    def get_token(self) -> str:
        """Get a valid JWT token, refreshing if necessary.

        Returns:
            Valid JWT access token

        Raises:
            Exception: If authentication fails
        """
        with self._lock:
            if self._is_token_valid():
                return cast(str, self._token)

            # Need to authenticate or refresh
            log.debug("Authenticating with pulsar-relay at %s", self.relay_url)
            self._authenticate()
            return cast(str, self._token)

    def _is_token_valid(self) -> bool:
        """Check if current token is valid and not expiring soon.

        Returns:
            True if token exists and won't expire soon, False otherwise
        """
        if self._token is None or self._token_expiry is None:
            return False

        # Check if token will expire within refresh buffer
        time_until_expiry = (self._token_expiry - datetime.now()).total_seconds()
        return time_until_expiry > self._refresh_buffer_seconds

    def _authenticate(self) -> None:
        """Perform authentication and cache the token.

        Raises:
            Exception: If authentication fails
        """

        auth_url = f"{self.relay_url}/auth/login"

        try:
            response = requests.post(
                auth_url,
                data={
                    'username': self.username,
                    'password': self.password,
                    'grant_type': 'password'
                },
                headers={'Content-Type': 'application/x-www-form-urlencoded'},
                timeout=10
            )
            response.raise_for_status()

            data = response.json()
            self._token = data['access_token']
            expires_in = data['expires_in']

            # Calculate expiry time
            self._token_expiry = datetime.now() + timedelta(seconds=expires_in)

            log.info("Successfully authenticated with pulsar-relay, token expires in %d seconds", expires_in)

        except requests.RequestException as e:
            log.error("Failed to authenticate with pulsar-relay: %s", e)
            raise Exception(f"pulsar-relay authentication failed: {e}")

    def invalidate(self) -> None:
        """Invalidate the current token, forcing re-authentication on next request."""
        with self._lock:
            self._token = None
            self._token_expiry = None
            log.debug("Invalidated pulsar-relay authentication token")



import logging

try:
    from urllib2 import urlopen
except ImportError:
    from urllib.request import urlopen
try:
    from urllib2 import Request
except ImportError:
    from urllib.request import Request

try:
    import poster
except ImportError:
    poster = None

POSTER_UNAVAILABLE_MESSAGE = "Pulsar configured to use poster module - but it is unavailable. Please install poster."

log = logging.getLogger(__name__)


if poster is not None:
    poster.streaminghttp.register_openers()


def post_file(url, path):
    __ensure_poster()
    try:
        datagen, headers = poster.encode.multipart_encode({"file": open(path, "rb")})
        request = Request(url, datagen, headers)
        return urlopen(request).read()
    except Exception:
        log.exception("Problem with poster post of [%s]" % path)
        raise


def get_file(url, path):
    __ensure_poster()
    request = Request(url=url)
    response = urlopen(request)
    with open(path, 'wb') as output:
        while True:
            buffer = response.read(1024)
            if not buffer:
                break
            output.write(buffer)


def __ensure_poster():
    if poster is None:
        raise ImportError(POSTER_UNAVAILABLE_MESSAGE)



import hashlib
import json
import os.path
import shutil
from base64 import (
    b64decode as _b64decode,
    b64encode as _b64encode,
)
from enum import Enum
from errno import (
    EEXIST,
    ENOENT,
)
from functools import wraps
from os import (
    curdir,
    listdir,
    makedirs,
    unlink,
    walk,
)
from os.path import (
    abspath,
    exists,
    join,
    relpath,
)
from threading import (
    Event,
    Lock,
)
from weakref import WeakValueDictionary

# TODO: move to galaxy.util so it doesn't have to be duplicated
# twice in pulsar.
BUFFER_SIZE = 4096


def copy_to_path(object, path):
    """
    Copy file-like object to path.
    """
    output = open(path, 'wb')
    _copy_and_close(object, output)


def _copy_and_close(object, output):
    try:
        while True:
            buffer = object.read(BUFFER_SIZE)
            if not buffer:
                break
            output.write(buffer)
    finally:
        output.close()


# Variant of base64 compat layer inspired by BSD code from Bcfg2
# https://github.com/Bcfg2/bcfg2/blob/maint/src/lib/Bcfg2/Compat.py
@wraps(_b64encode)
def b64encode(val, **kwargs):
    try:
        return _b64encode(val, **kwargs).decode("utf-8")
    except TypeError:
        return _b64encode(val.encode('UTF-8'), **kwargs).decode('UTF-8')


@wraps(_b64decode)
def b64decode(val, **kwargs):
    return _b64decode(val.encode('UTF-8'), **kwargs).decode('UTF-8')


def unique_path_prefix(path):
    m = hashlib.md5()
    m.update(path.encode('utf-8'))
    return m.hexdigest()


def copy(source, destination):
    """ Copy file from source to destination if needed (skip if source
    is destination).
    """
    source = os.path.abspath(source)
    destination = os.path.abspath(destination)
    if source != destination:
        if not os.path.exists(os.path.dirname(destination)):
            os.makedirs(os.path.dirname(destination))
        shutil.copyfile(source, destination)


def ensure_directory(file_path):
    directory = os.path.dirname(file_path)
    if not os.path.exists(directory):
        os.makedirs(directory)


def directory_files(directory):
    """

    >>> from tempfile import mkdtemp
    >>> from shutil import rmtree
    >>> from os.path import join
    >>> from os import makedirs
    >>> tempdir = mkdtemp()
    >>> with open(join(tempdir, "moo"), "w") as f: pass
    >>> directory_files(tempdir)
    ['moo']
    >>> subdir = join(tempdir, "cow", "sub1")
    >>> makedirs(subdir)
    >>> with open(join(subdir, "subfile1"), "w") as f: pass
    >>> with open(join(subdir, "subfile2"), "w") as f: pass
    >>> sorted(directory_files(tempdir))
    ['cow/sub1/subfile1', 'cow/sub1/subfile2', 'moo']
    >>> rmtree(tempdir)
    """
    contents = []
    for path, _, files in walk(directory):
        relative_path = relpath(path, directory)
        for name in files:
            # Return file1.txt, dataset_1_files/image.png, etc... don't
            # include . in path.
            if relative_path != curdir:
                contents.append(join(relative_path, name))
            else:
                contents.append(name)
    return contents


def filter_destination_params(destination_params, prefix):
    destination_params = destination_params or {}
    return {
        key[len(prefix):]: destination_params[key]
        for key in destination_params if key.startswith(prefix)
    }


def to_base64_json(data):
    """

    >>> enc = to_base64_json(dict(a=5))
    >>> dec = from_base64_json(enc)
    >>> dec["a"]
    5
    """
    dumped = json_dumps(data)
    return b64encode(dumped)


def from_base64_json(data):
    return json.loads(b64decode(data))


class PathHelper:
    '''

    >>> import posixpath
    >>> # Forcing local path to posixpath because Pulsar designed to be used with
    >>> # posix client.
    >>> posix_path_helper = PathHelper("/", local_path_module=posixpath)
    >>> windows_slash = "\\\\"
    >>> len(windows_slash)
    1
    >>> nt_path_helper = PathHelper(windows_slash, local_path_module=posixpath)
    >>> posix_path_helper.remote_name("moo/cow")
    'moo/cow'
    >>> nt_path_helper.remote_name("moo/cow")
    'moo\\\\cow'
    >>> posix_path_helper.local_name("moo/cow")
    'moo/cow'
    >>> nt_path_helper.local_name("moo\\\\cow")
    'moo/cow'
    >>> posix_path_helper.from_posix_with_new_base("/galaxy/data/bowtie/hg19.fa", "/galaxy/data/", "/work/galaxy/data")
    '/work/galaxy/data/bowtie/hg19.fa'
    >>> posix_path_helper.from_posix_with_new_base("/galaxy/data/bowtie/hg19.fa", "/galaxy/data", "/work/galaxy/data")
    '/work/galaxy/data/bowtie/hg19.fa'
    >>> posix_path_helper.from_posix_with_new_base("/galaxy/data/bowtie/hg19.fa", "/galaxy/data", "/work/galaxy/data/")
    '/work/galaxy/data/bowtie/hg19.fa'
    '''

    def __init__(self, separator, local_path_module=os.path):
        self.separator = separator
        self.local_join = local_path_module.join
        self.local_sep = local_path_module.sep

    def remote_name(self, local_name):
        return self.remote_join(*local_name.split(self.local_sep))

    def local_name(self, remote_name):
        return self.local_join(*remote_name.split(self.separator))

    def remote_join(self, *args):
        return self.separator.join(args)

    def from_posix_with_new_base(self, posix_path, old_base, new_base):
        # TODO: Test with new_base as a windows path against nt_path_helper.
        if old_base.endswith("/"):
            old_base = old_base[:-1]
        if not posix_path.startswith(old_base):
            message_template = "Cannot compute new path for file %s, does not start with %s."
            message = message_template % (posix_path, old_base)
            raise Exception(message)
        stripped_path = posix_path[len(old_base):]
        while stripped_path.startswith("/"):
            stripped_path = stripped_path[1:]
        path_parts = stripped_path.split(self.separator)
        if new_base.endswith(self.separator):
            new_base = new_base[:-len(self.separator)]
        return self.remote_join(new_base, *path_parts)


class TransferEventManager:

    def __init__(self):
        self.events = WeakValueDictionary(dict())
        self.events_lock = Lock()

    def acquire_event(self, path, force_clear=False):
        with self.events_lock:
            if path in self.events:
                event_holder = self.events[path]
            else:
                event_holder = EventHolder(Event(), path, self)
                self.events[path] = event_holder
        if force_clear:
            event_holder.event.clear()
        return event_holder


class EventHolder:

    def __init__(self, event, path, condition_manager):
        self.event = event
        self.path = path
        self.condition_manager = condition_manager
        self.failed = False

    def release(self):
        self.event.set()

    def fail(self):
        self.failed = True


def json_loads(obj):
    if isinstance(obj, bytes):
        obj = obj.decode("utf-8")
    return json.loads(obj)


def json_dumps(obj):
    if isinstance(obj, bytes):
        obj = obj.decode("utf-8")
    return json.dumps(obj, cls=ClientJsonEncoder)


class ClientJsonEncoder(json.JSONEncoder):

    def default(self, obj):
        if isinstance(obj, bytes):
            return obj.decode("utf-8")
        return json.JSONEncoder.default(self, obj)


class MessageQueueUUIDStore:
    """Persistent dict-like object for persisting message queue UUIDs that are
    awaiting acknowledgement or that have been operated on.
    """

    def __init__(self, persistence_directory, subdirs=None):
        if subdirs is None:
            subdirs = ['acknowledge_uuids']
        self.__store = abspath(join(persistence_directory, *subdirs))
        try:
            makedirs(self.__store)
        except OSError as exc:
            if exc.errno != EEXIST:
                raise

    def __path(self, item):
        return join(self.__store, item)

    def __contains__(self, item):
        return exists(self.__path(item))

    def __setitem__(self, key, value):
        open(self.__path(key), 'w').write(json.dumps(value))

    def __getitem__(self, key):
        return json.loads(open(self.__path(key)).read())

    def __delitem__(self, key):
        try:
            unlink(self.__path(key))
        except OSError as exc:
            if exc.errno == ENOENT:
                raise KeyError(key)
            raise

    def keys(self):
        return iter(listdir(self.__store))

    def get_time(self, key):
        try:
            return os.stat(self.__path(key)).st_mtime
        except OSError as exc:
            if exc.errno == ENOENT:
                raise KeyError(key)
            raise

    def set_time(self, key):
        try:
            os.utime(self.__path(key), None)
        except OSError as exc:
            if exc.errno == ENOENT:
                raise KeyError(key)
            raise


class ExternalId:
    external_id: str

    def __init__(self, external_id: str):
        self.external_id = external_id


class MonitorStyle(str, Enum):
    FOREGROUND = "foreground"
    BACKGROUND = "background"
    NONE = "none"



"""
"""
import os.path
import posixpath
from collections import deque
from glob import glob
from logging import getLogger

from galaxy.util import in_directory

from .util import PathHelper

log = getLogger(__name__)


TYPES_TO_METHOD = dict(
    input="inputs_directory",
    unstructured="unstructured_files_directory",
    config="configs_directory",
    tool="tool_files_directory",
    jobdir="job_directory",
    workdir="working_directory",
    metadata="metadata_directory",
    output="outputs_directory",
    output_workdir="working_directory",
    output_metadata="metadata_directory",
    output_jobdir="job_directory",
)


class RemoteJobDirectory:
    """ Representation of a (potentially) remote Pulsar-style staging directory.
    """

    def __init__(self, remote_staging_directory, remote_id, remote_sep):
        self.path_helper = PathHelper(remote_sep)
        if remote_id:
            self.job_directory = self.path_helper.remote_join(
                remote_staging_directory,
                remote_id
            )
        else:
            self.job_directory = remote_staging_directory

    def home_directory(self):
        return self._sub_dir('home')

    def metadata_directory(self):
        return self._sub_dir('metadata')

    def working_directory(self):
        return self._sub_dir('working')

    def inputs_directory(self):
        return self._sub_dir('inputs')

    def outputs_directory(self):
        return self._sub_dir('outputs')

    def configs_directory(self):
        return self._sub_dir('configs')

    def tool_files_directory(self):
        return self._sub_dir('tool_files')

    def unstructured_files_directory(self):
        return self._sub_dir('unstructured')

    def default_tmp_directory(self):
        return self._sub_dir('tmp')

    @property
    def path(self):
        return self.job_directory

    @property
    def separator(self):
        return self.path_helper.separator

    def calculate_path(self, remote_relative_path, input_type):
        """ Only for used by Pulsar client, should override for managers to
        enforce security and make the directory if needed.
        """
        directory, allow_nested_files, allow_globs = self._directory_for_file_type(input_type)
        return self.path_helper.remote_join(directory, remote_relative_path)

    def _directory_for_file_type(self, file_type):
        allow_nested_files = False
        # work_dir and input_extra are types used by legacy clients...
        # Obviously this client won't be legacy because this is in the
        # client module, but this code is reused on server which may
        # serve legacy clients.
        allow_nested_files = file_type in ['input', 'unstructured', 'output', 'output_workdir', 'metadata', 'output_metadata', 'tool']
        allow_globs = file_type in ['output_workdir']  # TODO: tool profile version where this is invalid
        directory_source = getattr(self, TYPES_TO_METHOD.get(file_type, None), None)
        if not directory_source:
            raise Exception("Unknown file_type specified %s" % file_type)
        if callable(directory_source):
            directory_source = directory_source()
        return directory_source, allow_nested_files, allow_globs

    def _sub_dir(self, name):
        return self.path_helper.remote_join(self.job_directory, name)


def get_mapped_file(directory, remote_path, allow_nested_files=False, local_path_module=os.path, mkdir=True, allow_globs=False):
    """

    >>> import ntpath
    >>> get_mapped_file(r'C:\\pulsar\\staging\\101', 'dataset_1_files/moo/cow', allow_nested_files=True, local_path_module=ntpath, mkdir=False)
    'C:\\\\pulsar\\\\staging\\\\101\\\\dataset_1_files\\\\moo\\\\cow'
    >>> get_mapped_file(r'C:\\pulsar\\staging\\101', 'dataset_1_files/moo/cow', allow_nested_files=False, local_path_module=ntpath)
    'C:\\\\pulsar\\\\staging\\\\101\\\\cow'
    >>> get_mapped_file(r'C:\\pulsar\\staging\\101', '../cow', allow_nested_files=True, local_path_module=ntpath, mkdir=False)
    Traceback (most recent call last):
    Exception: Attempt to read or write file outside an authorized directory.
    """
    if not allow_nested_files:
        name = local_path_module.basename(remote_path)
        path = local_path_module.join(directory, name)
    else:
        local_rel_path = __posix_to_local_path(remote_path, local_path_module=local_path_module)
        local_path = local_path_module.join(directory, local_rel_path)
        verify_is_in_directory(local_path, directory, local_path_module=local_path_module)
        local_directory = local_path_module.dirname(local_path)
        if mkdir and not local_path_module.exists(local_directory):
            os.makedirs(local_directory)
        path = local_path
    if allow_globs and ('*' in path or '?' in path):
        matches = glob(path)
        if len(matches) == 0:
            raise RuntimeError(f"No files matching glob: {path}")
        elif len(matches) > 1:
            log.warning(f"Found multiple files matching {path}, using the first match: {matches}")
        else:
            log.info(f"Glob path {path} mapped to matched file: {matches[0]}")
        path = matches[0]
    return path


def __posix_to_local_path(path, local_path_module=os.path):
    """
    Converts a posix path (coming from Galaxy), to a local path (be it posix or Windows).

    >>> import ntpath
    >>> __posix_to_local_path('dataset_1_files/moo/cow', local_path_module=ntpath)
    'dataset_1_files\\\\moo\\\\cow'
    >>> import posixpath
    >>> __posix_to_local_path('dataset_1_files/moo/cow', local_path_module=posixpath)
    'dataset_1_files/moo/cow'
    """
    partial_path = deque()
    while True:
        if not path or path == '/':
            break
        (path, base) = posixpath.split(path)
        partial_path.appendleft(base)
    return local_path_module.join(*partial_path)


def verify_is_in_directory(path, directory, local_path_module=os.path):
    if not in_directory(path, directory, local_path_module):
        msg = "Attempt to read or write file outside an authorized directory."
        log.warn("{} Attempted path: {}, valid directory: {}".format(msg, path, directory))
        raise Exception(msg)



"""Stand-alone entry point for running Pulsar without a web server.

In its simplest form, this method will check the current directory for an
app.yml and run the corresponding configuration as a standalone applciation.
This makes sense when ``app.yml`` contains a ``message_queue_url`` option so
Pulsar is configured to listen to a message queue and doesn't require a web
server.

The following commands can be used to bootstrap such a setup.::

    mkdir pulsar-mq-config
    cd pulsar-mq-config
    pulsar-config --mq
    pulsar-main

This script can be used in a standalone fashion, but it is generally better to
run the ``pulsar`` script with ``--mode webless`` - which will in turn
delegate to this script.
"""
import configparser
import functools
import logging
import os
import sys
import time
from logging.config import fileConfig

try:
    import yaml
except ImportError:
    yaml = None  # type: ignore

try:
    from daemonize import Daemonize
except ImportError:
    Daemonize = None

from argparse import (
    ArgumentParser,
    RawDescriptionHelpFormatter,
)

log = logging.getLogger(__name__)

REQUIRES_DAEMONIZE_MESSAGE = "Attempted to use Pulsar in daemon mode, but daemonize is unavailable."

PULSAR_ROOT_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
if "PULSAR_CONFIG_DIR" in os.environ:
    PULSAR_CONFIG_DIR = os.path.abspath(os.environ["PULSAR_CONFIG_DIR"])
else:
    PULSAR_CONFIG_DIR = PULSAR_ROOT_DIR

DEFAULT_INI_APP = "main"
DEFAULT_INI = "server.ini"
DEFAULT_APP_YAML = "app.yml"
DEFAULT_MANAGER = "_default_"

DEFAULT_PID = "pulsar.pid"
DEFAULT_VERBOSE = True
HELP_CONFIG_DIR = "Default directory to search for relevant Pulsar configuration files (e.g. app.yml, server.ini)."
HELP_INI_PATH = "Specify an explicit path to Pulsar's server.ini configuration file."
HELP_APP_CONF_PATH = "Specify an explicit path to Pulsar's app.yml configuration file."
HELP_APP_CONF_BASE64 = "Specify an application configuration as a base64 encoded JSON blob."
HELP_DAEMONIZE = "Daemonzie process (requires daemonize library)."
CONFIG_PREFIX = "PULSAR_CONFIG_"


LOGGING_CONFIG_DEFAULT = {
    'version': 1,
    'root': {
        'handlers': ['console'],
        'level': 'INFO',
    },
    'loggers': {
        'pulsar': {
            'handlers': ['console'],
            'level': 'DEBUG',
            'propagate': 0,
            'qualname': 'pulsar',
        },
        'galaxy': {
            'handlers': ['console'],
            'level': 'DEBUG',
            'propagate': 0,
            'qualname': 'pulsar',
        },
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'formatter': 'default',
            'level': 'DEBUG',
            'stream': 'ext://sys.stderr',
        },
    },
    'formatters': {
        'default': {
            'format': '%(asctime)s %(levelname)-5.5s [%(name)s][%(threadName)s] %(message)s'
        },
    },
}


def load_pulsar_app(
    config_builder,
    config_env=False,
    log=None,
    **kwds
):
    # Allow specification of log so daemon can reuse properly configured one.
    if log is None:
        log = logging.getLogger(__name__)

    # If called in daemon mode, set the ROOT directory and ensure Pulsar is on
    # sys.path.
    if config_env:
        try:
            os.chdir(PULSAR_ROOT_DIR)
        except Exception:
            log.exception("Failed to chdir")
            raise
        try:
            sys.path.append(PULSAR_ROOT_DIR)
        except Exception:
            log.exception("Failed to add Pulsar to sys.path")
            raise

    config_builder.setup_file_logging()
    config = config_builder.load()

    config.update(kwds)
    import pulsar.core
    pulsar_app = pulsar.core.PulsarApp(**config)
    return pulsar_app


def app_loop(args, log, config_env):
    pulsar_app = _app(args, log, config_env)
    sleep = True
    while sleep:
        try:
            time.sleep(5)
        except KeyboardInterrupt:
            sleep = False
        except SystemExit:
            sleep = False
        except Exception:
            pass
    try:
        pulsar_app.shutdown()
    except Exception:
        log.exception("Failed to shutdown Pulsar application")
        raise


def _app(args, log, config_env):
    try:
        config_builder = PulsarConfigBuilder(args)
        pulsar_app = load_pulsar_app(
            config_builder,
            config_env=config_env,
            log=log,
        )
    except BaseException:
        log.exception("Failed to initialize Pulsar application")
        raise
    return pulsar_app


def absolute_config_path(path, config_dir):
    if path and not os.path.isabs(path):
        path = os.path.join(config_dir, path)
    return path


def _find_default_app_config(*config_dirs):
    for config_dir in config_dirs:
        app_config_path = os.path.join(config_dir, DEFAULT_APP_YAML)
        if os.path.exists(app_config_path):
            return app_config_path
    return None


def apply_env_overrides_and_defaults(conf):
    override_prefix = "%sOVERRIDE_" % CONFIG_PREFIX
    for key in os.environ:
        if key == 'PULSAR_CONFIG_DIR':
            conf['config_dir'] = os.environ[key]
        elif key.startswith(override_prefix):
            config_key = key[len(override_prefix):].lower()
            conf[config_key] = os.environ[key]
        elif key.startswith(CONFIG_PREFIX):
            config_key = key[len(CONFIG_PREFIX):].lower()
            if config_key not in conf:
                conf[config_key] = os.environ[key]
    return conf


def load_app_configuration(ini_path=None, app_conf_path=None, app_name=None, local_conf=None, config_dir=PULSAR_CONFIG_DIR):
    """
    """
    if ini_path and local_conf is None:
        # Simple ini parsing without loadwsgi dependencies
        ini_path_abs = os.path.abspath(ini_path)
        defaults = {
            'here': os.path.dirname(ini_path_abs),
            '__file__': ini_path_abs
        }
        config = configparser.ConfigParser(defaults=defaults)
        config.read(ini_path)
        section_name = f"app:{app_name}"
        if config.has_section(section_name):
            local_conf = dict(config.items(section_name))
        else:
            local_conf = {}
    local_conf = local_conf or {}
    local_conf['config_dir'] = config_dir
    if app_conf_path is None and "app_config" in local_conf:
        app_conf_path = absolute_config_path(local_conf["app_config"], config_dir)
        if not os.path.exists(app_conf_path) and os.path.exists(app_conf_path + ".sample"):
            app_conf_path = app_conf_path + ".sample"
    elif ini_path:
        # If not explicit app.yml file found - look next to server.ini -
        # be it in pulsar root, some temporary staging directory, or /etc.
        app_conf_path = _find_default_app_config(
            os.path.dirname(ini_path),
        )
    if app_conf_path:
        if yaml is None:
            raise Exception("Cannot load configuration from file %s, pyyaml is not available." % app_conf_path)

        with open(app_conf_path) as f:
            app_conf = yaml.safe_load(f) or {}
            local_conf.update(app_conf)

    return apply_env_overrides_and_defaults(local_conf)


def find_ini(supplied_ini, config_dir):
    if supplied_ini:
        return supplied_ini

    # If not explicitly supplied an ini, check server.ini and then
    # just resort to sample if that has not been configured.
    for guess in ["server.ini", "server.ini.sample"]:
        ini_path = os.path.join(config_dir, guess)
        if os.path.exists(ini_path):
            return ini_path

    return guess


class PulsarConfigBuilder:
    """ Generate paste-like configuration from supplied command-line arguments.
    """

    def __init__(self, args=None, **kwds):
        config_dir = kwds.get("config_dir", None) or (args and args.config_dir) or PULSAR_CONFIG_DIR
        ini_path = kwds.get("ini_path", None) or (args and args.ini_path)
        app_conf_path = kwds.get("app_conf_path", None) or (args and args.app_conf_path)
        app_conf_base64 = args and args.app_conf_base64

        if not app_conf_base64 and not app_conf_path:
            # If given app_conf_path - use that - else we need to ensure we have an
            # ini path.
            ini_path = find_ini(ini_path, config_dir)
            ini_path = absolute_config_path(ini_path, config_dir=config_dir)
        self.config_dir = config_dir
        self.ini_path = ini_path
        self.app_conf_path = app_conf_path
        self.app_conf_base64 = app_conf_base64
        self.app_name = kwds.get("app") or (args and args.app) or DEFAULT_INI_APP

    @classmethod
    def populate_options(cls, arg_parser):
        arg_parser.add_argument("-c", "--config_dir", default=None, help=HELP_CONFIG_DIR)
        arg_parser.add_argument("--ini_path", default=None, help=HELP_INI_PATH)
        arg_parser.add_argument("--app_conf_path", default=None, help=HELP_APP_CONF_PATH)
        arg_parser.add_argument("--app_conf_base64", default=None, help=HELP_APP_CONF_BASE64)
        arg_parser.add_argument("--app", default=DEFAULT_INI_APP)
        # daemon related options...
        arg_parser.add_argument("-d", "--daemonize", default=False, help=HELP_DAEMONIZE, action="store_true")
        arg_parser.add_argument("--daemon-log-file", default=None, help="Log file for daemon, if --daemonize supplied.")
        arg_parser.add_argument("--pid-file", default=DEFAULT_PID, help="Pid file for daemon, if --daemonize supplied (default is %s)." % DEFAULT_PID)

    def load(self):
        load_kwds = dict(
            app_name=self.app_name,
            config_dir=self.config_dir,
        )
        if self.app_conf_base64:
            from pulsar.client.util import from_base64_json
            local_conf = from_base64_json(self.app_conf_base64)
            self.setup_dict_logging(local_conf)
            load_kwds["local_conf"] = local_conf
        else:
            load_kwds.update(dict(
                config_dir=self.config_dir,
                ini_path=self.ini_path,
                app_conf_path=self.app_conf_path,
            ))
        return load_app_configuration(**load_kwds)

    def setup_file_logging(self):
        if self.ini_path:
            raw_config = configparser.ConfigParser()
            raw_config.read([self.ini_path])
            # https://github.com/mozilla-services/chaussette/pull/32/files
            if raw_config.has_section('loggers'):
                config_file = os.path.abspath(self.ini_path)
                fileConfig(
                    config_file,
                    dict(__file__=config_file, here=os.path.dirname(config_file))
                )

    def setup_dict_logging(self, config):
        logging_conf = config.get('logging', None)
        if logging_conf is None:
            # if using the default logging config, honor the log_level setting
            logging_conf = LOGGING_CONFIG_DEFAULT
        logging.config.dictConfig(logging_conf)

    def to_dict(self):
        return dict(
            config_dir=self.config_dir,
            ini_path=self.ini_path,
            app_conf_path=self.app_conf_path,
            app=self.app_name
        )


class PulsarManagerConfigBuilder(PulsarConfigBuilder):

    def __init__(self, args=None, **kwds):
        super().__init__(args=args, **kwds)
        self.manager = kwds.get("manager", None) or (args and args.manager) or DEFAULT_MANAGER

    def to_dict(self):
        as_dict = super().to_dict()
        as_dict["manager"] = self.manager
        return as_dict

    @classmethod
    def populate_options(cls, arg_parser):
        PulsarConfigBuilder.populate_options(arg_parser)
        arg_parser.add_argument("--manager", default=DEFAULT_MANAGER)


def main(argv=None, config_env=False):
    mod_docstring = sys.modules[__name__].__doc__
    arg_parser = ArgumentParser(
        description=mod_docstring,
        formatter_class=RawDescriptionHelpFormatter,
    )
    PulsarConfigBuilder.populate_options(arg_parser)
    args = arg_parser.parse_args(argv)

    pid_file = args.pid_file

    log.setLevel(logging.DEBUG)
    log.propagate = False

    if args.daemonize:
        if Daemonize is None:
            raise ImportError(REQUIRES_DAEMONIZE_MESSAGE)

        keep_fds = []
        if args.daemon_log_file:
            fh = logging.FileHandler(args.daemon_log_file, "w")
            fh.setLevel(logging.DEBUG)
            log.addHandler(fh)
            keep_fds.append(fh.stream.fileno())
        else:
            fh = logging.StreamHandler(sys.stderr)
            fh.setLevel(logging.DEBUG)
            log.addHandler(fh)

        daemon = Daemonize(
            app="pulsar",
            pid=pid_file,
            action=functools.partial(app_loop, args, log, config_env),
            verbose=DEFAULT_VERBOSE,
            logger=log,
            keep_fds=keep_fds,
        )
        daemon.start()
    else:
        app_loop(args, log, config_env)


if __name__ == "__main__":
    main(config_env=True)



"""
"""
import logging
import os

from pulsar.client import (
    action_mapper,
    staging,
)
from pulsar.client.staging import PulsarOutputs
from pulsar.client.staging.down import ResultsCollector

log = logging.getLogger(__name__)


def postprocess(job_directory, action_executor, was_cancelled):
    # Returns True if outputs were collected.
    try:
        if job_directory.has_metadata("launch_config"):
            staging_config = job_directory.load_metadata("launch_config").get("remote_staging", None)
        else:
            staging_config = None
        collected = __collect_outputs(job_directory, staging_config, action_executor, was_cancelled)
        return collected
    finally:
        job_directory.write_file("postprocessed", "")
    return False


def __collect_outputs(job_directory, staging_config, action_executor, was_cancelled):
    collected = True
    if "action_mapper" in staging_config:
        file_action_mapper = action_mapper.FileActionMapper(config=staging_config["action_mapper"])
        client_outputs = staging.ClientOutputs.from_dict(staging_config["client_outputs"])
        pulsar_outputs = __pulsar_outputs(job_directory)
        output_collector = PulsarServerOutputCollector(job_directory, action_executor, was_cancelled)
        results_collector = ResultsCollector(output_collector, file_action_mapper, client_outputs, pulsar_outputs)
        collection_failure_exceptions = results_collector.collect()
        if collection_failure_exceptions:
            log.warn("Failures collecting results %s" % collection_failure_exceptions)
            collected = False
    return collected


def realized_dynamic_file_sources(job_directory):
    launch_config = job_directory.load_metadata("launch_config")
    if launch_config is None:
        log.warning(f"Failed to load launch_config from: {job_directory.job_directory}")
        return []
    dynamic_file_sources = launch_config.get("dynamic_file_sources")
    realized_dynamic_file_sources = []
    for dynamic_file_source in (dynamic_file_sources or []):
        dynamic_file_source_path = dynamic_file_source["path"]
        realized_dynamic_file_source = dynamic_file_source.copy()
        dynamic_file_source_bytes = job_directory.working_directory_file_contents(dynamic_file_source_path)
        if dynamic_file_source_bytes is not None:
            dynamic_file_source_contents = dynamic_file_source_bytes.decode("utf-8")
            realized_dynamic_file_source["contents"] = dynamic_file_source_contents
            realized_dynamic_file_sources.append(realized_dynamic_file_source)
    return realized_dynamic_file_sources


class PulsarServerOutputCollector:

    def __init__(self, job_directory, action_executor, was_cancelled):
        self.job_directory = job_directory
        self.action_executor = action_executor
        self.was_cancelled = was_cancelled

    def collect_output(self, results_collector, output_type, action, name):
        def action_if_not_cancelled():
            if self.was_cancelled():
                log.info(f"Skipped output collection '{name}', job is cancelled")
                return
            action.write_from_path(pulsar_path)
        # Not using input path, this is because action knows it path
        # in this context.
        if action.staging_action_local:
            return  # Galaxy (client) will collect output.

        if not name:
            # TODO: Would not work on Windows. Any use in allowing
            # remote_transfer action for Windows?
            name = os.path.basename(action.path)

        pulsar_path = self.job_directory.calculate_path(name, output_type)
        description = "staging out file {} via {}".format(pulsar_path, action)
        self.action_executor.execute(action_if_not_cancelled, description)


def __pulsar_outputs(job_directory):
    working_directory_contents = job_directory.working_directory_contents()
    output_directory_contents = job_directory.outputs_directory_contents()
    metadata_directory_contents = job_directory.metadata_directory_contents()
    job_directory_contents = job_directory.job_directory_contents()
    return PulsarOutputs(
        working_directory_contents,
        output_directory_contents,
        metadata_directory_contents,
        job_directory_contents,
        realized_dynamic_file_sources=realized_dynamic_file_sources(job_directory),
    )


__all__ = ('postprocess', 'realized_dynamic_file_sources')



import configparser
import inspect
import logging
import os

import pulsar.managers
from pulsar.managers import stateful

log = logging.getLogger(__name__)


MANAGER_PREFIX = 'manager:'
DEFAULT_MANAGER_NAME = '_default_'
DEFAULT_MANAGER_TYPE = 'queued_python'


def build_managers(app, conf):
    """
    Takes in a config file as outlined in job_managers.ini.sample and builds
    a dictionary of job manager objects from them.
    """
    # Load default options from config file that apply to all
    # managers.
    default_options = _get_default_options(conf)

    manager_descriptions = ManagerDescriptions()
    if "job_managers_config" in conf:
        job_managers_config = conf.get("job_managers_config", None)
        _populate_manager_descriptions_from_ini(manager_descriptions, job_managers_config)
    elif "managers" in conf:
        for manager_name, manager_options in conf["managers"].items():
            manager_description = ManagerDescription.from_dict(manager_options, manager_name)
            manager_descriptions.add(manager_description)
    elif "manager" in conf:
        manager_description = ManagerDescription.from_dict(conf["manager"])
        manager_descriptions.add(manager_description)
    else:
        manager_descriptions.add(ManagerDescription())

    manager_classes = _get_managers_dict()
    managers = {}
    for manager_name, manager_description in manager_descriptions.descriptions.items():
        manager_options = dict(default_options)
        manager_options.update(manager_description.manager_options)

        manager_class = manager_classes[manager_description.manager_type]
        manager = _build_manager(manager_class, app, manager_name, manager_options)
        managers[manager_name] = manager

    return managers


def _populate_manager_descriptions_from_ini(manager_descriptions, job_managers_config):
    config = configparser.ConfigParser()
    config.readfp(open(job_managers_config))
    for section in config.sections():
        if not section.startswith(MANAGER_PREFIX):
            continue
        manager_name = section[len(MANAGER_PREFIX):]
        manager_description = ManagerDescription.from_ini_config(config, manager_name)
        manager_descriptions.add(manager_description)


def _get_default_options(conf):
    options = {}
    for simple_key in ["assign_ids", "galaxy_home"]:
        if simple_key in conf:
            options[simple_key] = conf[simple_key]
    options["debug"] = conf.get("debug", False)
    maximum_stream_size = conf.get("maximum_stream_size", 1024 * 1024)
    if maximum_stream_size:
        options["maximum_stream_size"] = int(maximum_stream_size)
    # mode to create job directories with, if None just use
    # default (usually 0777 with umask applied).
    job_directory_mode = conf.get("job_directory_mode", None)
    options["job_directory_mode"] = None
    if job_directory_mode is not None:
        options["job_directory_mode"] = int(job_directory_mode, 8)
    return options


def _build_manager(manager_class, app, name=DEFAULT_MANAGER_NAME, manager_options={}):
    return stateful.StatefulManagerProxy(manager_class(name, app, **manager_options), **manager_options)


def _get_manager_modules():
    """

    >>> 'pulsar.managers.queued_pbs' in _get_manager_modules()
    True
    >>> 'pulsar.managers.queued_drmaa' in _get_manager_modules()
    True
    """
    managers_dir = pulsar.managers.__path__[0]
    module_names = []
    for fname in os.listdir(managers_dir):
        if not fname.startswith("_") and fname.endswith(".py"):
            manager_module_name = "pulsar.managers.%s" % fname[:-len(".py")]
            module_names.append(manager_module_name)
    return module_names


def _load_manager_modules():
    modules = []
    for manager_module_name in _get_manager_modules():
        try:
            module = __import__(manager_module_name)
            for comp in manager_module_name.split(".")[1:]:
                module = getattr(module, comp)
            modules.append(module)
        except BaseException as exception:
            exception_str = str(exception)
            message = "{} manager module could not be loaded: {}".format(manager_module_name, exception_str)
            log.warn(message)
            continue

    return modules


def _get_managers_dict():
    """

    >>> from pulsar.managers.queued_pbs import PbsQueueManager
    >>> _get_managers_dict()['queued_pbs'] == PbsQueueManager
    True
    >>> from pulsar.managers.queued_drmaa import DrmaaQueueManager
    >>> _get_managers_dict()['queued_drmaa'] == DrmaaQueueManager
    True
    """
    managers = {}
    for manager_module in _load_manager_modules():
        for _, obj in inspect.getmembers(manager_module):
            if inspect.isclass(obj) and hasattr(obj, 'manager_type'):
                managers[obj.manager_type] = obj

    return managers


class ManagerDescriptions:

    def __init__(self):
        self.descriptions = {}

    def add(self, manager_description):
        manager_name = manager_description.manager_name
        if manager_name in self.descriptions:
            raise Exception("Problem configuring job managers, multiple managers with name %s" % manager_name)
        self.descriptions[manager_name] = manager_description


class ManagerDescription:

    def __init__(self, manager_type=DEFAULT_MANAGER_TYPE, manager_name=DEFAULT_MANAGER_NAME, manager_options={}):
        self.manager_type = manager_type
        self.manager_name = manager_name
        self.manager_options = manager_options

    @staticmethod
    def from_ini_config(config, manager_name):
        section_name = '{}{}'.format(MANAGER_PREFIX, manager_name)
        try:
            manager_type = config.get(section_name, 'type')
        except ValueError:
            manager_type = DEFAULT_MANAGER_TYPE

        # Merge default and specific manager options.
        manager_options = {}
        manager_options.update(dict(config.items(section_name)))
        return ManagerDescription(manager_type, manager_name, manager_options)

    @staticmethod
    def from_dict(config, manager_name=None):
        manager_type = config.get("type", DEFAULT_MANAGER_TYPE)
        manager_name = manager_name or config.get("name") or DEFAULT_MANAGER_NAME
        return ManagerDescription(manager_type, manager_name, config)



"""
HTTP transport for communicating with pulsar-relay.

Provides methods for posting messages, long-polling, and managing
authentication with the relay server.
"""
import logging
import time
from typing import Any, Callable, Dict, List, Optional

import requests

from ..relay_auth import RelayAuthManager

log = logging.getLogger(__name__)


class RelayTransportError(Exception):
    """Raised when communication with pulsar-relay fails."""
    pass


class RelayTransport:
    """HTTP transport for pulsar-relay communication.

    Handles:
    - Message publishing (single and bulk)
    - Long-polling for message consumption
    - Automatic authentication and retry
    """

    def __init__(self, relay_url: str, username: str, password: str, timeout: int = 30):
        """Initialize the relay transport.

        Args:
            relay_url: Base URL of the pulsar-relay server
            username: Username for authentication
            password: Password for authentication
            timeout: Default request timeout in seconds
        """
        self.relay_url = relay_url.rstrip('/')
        self.auth_manager = RelayAuthManager(relay_url, username, password)
        self.timeout = timeout
        self.session = requests.Session()
        self._last_message_ids: Dict[str, str] = {}  # Track last message ID per topic

    def _get_headers(self) -> Dict[str, str]:
        """Get HTTP headers including authentication token.

        Returns:
            Dictionary of HTTP headers
        """
        token = self.auth_manager.get_token()
        return {
            'Authorization': f'Bearer {token}',
            'Content-Type': 'application/json'
        }

    def _retry_with_backoff(
        self,
        operation: Callable[[], requests.Response],
        operation_name: str,
        initial_delay: float = 1.0,
        max_delay: float = 60.0,
        backoff_factor: float = 2.0
    ) -> requests.Response:
        """Retry an operation with exponential backoff.

        Retries indefinitely for communication errors and 5xx errors,
        but not for 4xx errors (client errors).

        Args:
            operation: Function that performs the HTTP request
            operation_name: Name of the operation for logging
            initial_delay: Initial delay in seconds before first retry
            max_delay: Maximum delay in seconds between retries
            backoff_factor: Multiplier for exponential backoff

        Returns:
            requests.Response object from successful request

        Raises:
            RelayTransportError: For 4xx client errors (non-retryable)
        """
        attempt = 0
        delay = initial_delay

        while True:
            attempt += 1
            try:
                response = operation()

                # Check for client errors (4xx) - don't retry these
                if 400 <= response.status_code < 500:
                    # Let caller handle 401 separately
                    if response.status_code == 401:
                        return response
                    # Other 4xx errors are client errors, don't retry
                    response.raise_for_status()
                    return response

                # Check for server errors (5xx) - retry these
                if response.status_code >= 500:
                    log.warning(
                        "%s failed with status %d (attempt %d), retrying in %.1f seconds",
                        operation_name, response.status_code, attempt, delay
                    )
                    time.sleep(delay)
                    delay = min(delay * backoff_factor, max_delay)
                    continue

                # Success
                return response

            except requests.ConnectionError as e:
                log.warning(
                    "%s connection error (attempt %d): %s, retrying in %.1f seconds",
                    operation_name, attempt, e, delay
                )
                time.sleep(delay)
                delay = min(delay * backoff_factor, max_delay)

            except requests.Timeout as e:
                log.warning(
                    "%s timeout (attempt %d): %s, retrying in %.1f seconds",
                    operation_name, attempt, e, delay
                )
                time.sleep(delay)
                delay = min(delay * backoff_factor, max_delay)

            except requests.RequestException as e:
                # For other request exceptions, check if it's a client error
                if hasattr(e, 'response') and e.response is not None:
                    if 400 <= e.response.status_code < 500:
                        # Client error, don't retry
                        raise RelayTransportError(f"{operation_name} failed: {e}")
                # Otherwise retry
                log.warning(
                    "%s error (attempt %d): %s, retrying in %.1f seconds",
                    operation_name, attempt, e, delay
                )
                time.sleep(delay)
                delay = min(delay * backoff_factor, max_delay)

    def post_message(
        self,
        topic: str,
        payload: Dict[str, Any],
        ttl: Optional[int] = None,
        metadata: Optional[Dict[str, str]] = None
    ) -> Dict[str, Any]:
        """Post a single message to the relay.

        Automatically retries on communication errors and server errors (5xx)
        with exponential backoff. Does not retry on client errors (4xx).

        Args:
            topic: Topic name to publish to
            payload: Message payload (must be JSON-serializable)
            ttl: Time-to-live in seconds (optional)
            metadata: Optional metadata dictionary

        Returns:
            Response dictionary with message_id, topic, and timestamp

        Raises:
            RelayTransportError: If the request fails with a client error (4xx)
        """
        url = f"{self.relay_url}/api/v1/messages"

        message_data: Dict[str, Any] = {
            'topic': topic,
            'payload': payload
        }

        if ttl is not None:
            message_data['ttl'] = ttl

        if metadata is not None:
            message_data['metadata'] = metadata

        def _post() -> requests.Response:
            return self.session.post(
                url,
                json=message_data,
                headers=self._get_headers(),
                timeout=self.timeout
            )

        # Use retry logic with exponential backoff
        response = self._retry_with_backoff(_post, "post_message")

        if response.status_code == 401:
            # Token might have expired, invalidate and retry
            log.debug("Received 401, invalidating token and retrying")
            self.auth_manager.invalidate()
            response = self._retry_with_backoff(_post, "post_message")

        response.raise_for_status()
        result = response.json()

        log.debug("Posted message to topic '%s': message_id=%s", topic, result.get('message_id'))
        return result

    def post_bulk_messages(self, messages: List[Dict[str, Any]]) -> Dict[str, Any]:
        """Post multiple messages in a single request.

        Automatically retries on communication errors and server errors (5xx)
        with exponential backoff. Does not retry on client errors (4xx).

        Args:
            messages: List of message dictionaries, each containing 'topic' and 'payload'

        Returns:
            Response dictionary with results and summary

        Raises:
            RelayTransportError: If the request fails with a client error (4xx)
        """
        url = f"{self.relay_url}/api/v1/messages/bulk"

        request_data = {'messages': messages}

        def _post() -> requests.Response:
            return self.session.post(
                url,
                json=request_data,
                headers=self._get_headers(),
                timeout=self.timeout
            )

        # Use retry logic with exponential backoff
        response = self._retry_with_backoff(_post, "post_bulk_messages")

        if response.status_code == 401:
            self.auth_manager.invalidate()
            response = self._retry_with_backoff(_post, "post_bulk_messages")

        response.raise_for_status()
        result = response.json()

        log.debug("Posted %d messages in bulk", len(messages))
        return result

    def long_poll(
        self,
        topics: List[str],
        timeout: int = 30,
    ) -> List[Dict[str, Any]]:
        """Poll for messages from specified topics.

        This is a blocking call that waits up to 'timeout' seconds for new messages.
        Automatically tracks the last message ID per topic and includes it in
        subsequent poll requests to ensure no messages are missed.

        Args:
            topics: List of topic names to subscribe to
            timeout: Maximum seconds to wait for messages (1-60)

        Returns:
            List of message dictionaries

        Raises:
            RelayTransportError: If the request fails
        """
        url = f"{self.relay_url}/messages/poll"

        poll_data = {
            'topics': topics,
            'timeout': min(max(timeout, 1), 60)  # Clamp to 1-60 range
        }

        # Build since dict from tracked message IDs for requested topics
        tracked_since = {}
        for topic in topics:
            msg_id = self.get_last_message_id(topic)
            if msg_id is not None:
                tracked_since[topic] = msg_id
        if tracked_since:
            poll_data['since'] = tracked_since
            log.debug("Using tracked message IDs for poll: %s", tracked_since)

        try:
            response = self.session.post(
                url,
                json=poll_data,
                headers=self._get_headers(),
                timeout=timeout + 5  # Add buffer to request timeout
            )

            if response.status_code == 401:
                self.auth_manager.invalidate()
                response = self.session.post(
                    url,
                    json=poll_data,
                    headers=self._get_headers(),
                    timeout=timeout + 5
                )

            response.raise_for_status()
            result = response.json()

            messages = result.get('messages') or []

            # Track the last message ID per topic
            for message in messages:
                topic = message.get('topic')
                message_id = message.get('message_id')
                if topic and message_id:
                    self.set_last_message_id(topic, message_id)

            # Build dict of tracked IDs for the requested topics
            tracked_ids = {}
            for topic in topics:
                msg_id = self.get_last_message_id(topic)
                if msg_id is not None:
                    tracked_ids[topic] = msg_id
            log.debug(
                "Received %d messages from long poll, tracked IDs: %s",
                len(messages),
                tracked_ids
            )

            return messages

        except requests.Timeout:
            # Timeout is expected in long polling when no messages arrive
            log.debug("Long poll timeout (no messages)")
            return []

        except requests.RequestException as e:
            log.error("Failed to long poll: %s", e)
            raise RelayTransportError(f"Failed to long poll: {e}")

    def get_last_message_id(self, topic: str) -> Optional[str]:
        """Get the last tracked message ID for a topic.

        Args:
            topic: Topic name

        Returns:
            Last message ID for the topic, or None if not tracked
        """
        return self._last_message_ids.get(topic)

    def get_all_tracked_message_ids(self) -> Dict[str, str]:
        """Get all tracked message IDs.

        Returns:
            Dictionary mapping topic names to last message IDs
        """
        return self._last_message_ids.copy()

    def set_last_message_id(self, topic: str, message_id: str) -> None:
        """Manually set the last message ID for a topic.

        This can be useful for resuming from a specific message ID
        after a restart.

        Args:
            topic: Topic name
            message_id: Message ID to set as the last seen message
        """
        self._last_message_ids[topic] = message_id

    def clear_tracked_message_ids(self, topic: Optional[str] = None) -> None:
        """Clear tracked message IDs.

        Args:
            topic: If provided, clears only the specified topic.
                  If None, clears all tracked message IDs.
        """
        if topic is not None:
            if topic in self._last_message_ids:
                del self._last_message_ids[topic]
                log.debug("Cleared tracked message ID for topic '%s'", topic)
        else:
            self._last_message_ids.clear()
            log.debug("Cleared all tracked message IDs")

    def close(self):
        """Close the transport and cleanup resources."""
        self.session.close()



import errno
import logging
import os
import signal
from time import sleep

log = logging.getLogger(__name__)


def check_pg(pgid):
    """Check whether processes in process group pgid are still alive."""
    try:
        (pid, exit_status) = os.waitpid(-pgid, os.WNOHANG)
    except OSError as e:
        if e.errno == errno.ECHILD:
            log.debug("check_pg(): No process found in process group %d", pgid)
        else:
            log.warning(
                "check_pg(): Got errno %s when checking process group %d: %s",
                errno.errorcode[e.errno],
                pgid,
                e.strerror,
            )
        return False
    # Since we are passing os.WNOHANG to os.waitpid(), pid is 0 if no process
    # status is available immediately.
    return pid == 0


def kill_pg(pgid):
    """Kill all processes in process group pgid."""
    for sig in [signal.SIGTERM, signal.SIGKILL]:
        try:
            os.killpg(pgid, sig)
        except OSError as e:
            if e.errno == errno.ESRCH:
                return
            log.warning(
                "Got errno %s when sending signal %d to process group %d: %s",
                errno.errorcode[e.errno],
                sig,
                pgid,
                e.strerror,
            )
        sleep(1)
        if not check_pg(pgid):
            log.debug("Processes in process group %d successfully killed with signal %d", pgid, sig)
            return
    else:
        log.warning("Some process in process group %d refuses to die after signaling TERM/KILL", pgid)



from pulsar.user_auth.methods.interface import AuthMethod


class UserListAuth(AuthMethod):
    """
    Defines authorization user by username
    """

    def __init__(self, config):
        try:
            self._allowed_users = config["userlist_allowed_users"]
        except Exception as e:
            raise Exception("cannot read UsernameAuth configuration") from e

    auth_type = "userlist"

    def authorize(self, authentication_info):
        return authentication_info["username"] in self._allowed_users

    def authenticate(self, job_directory):
        raise NotImplementedError("authentication not implemented for this class")



"""
"""
import json

from galaxy.util.plugin_config import plugins_dict

DEFAULT_SHELL_PLUGIN = "LocalShell"

ERROR_MESSAGE_NO_JOB_PLUGIN = "No job plugin parameter found, cannot create CLI job interface"
ERROR_MESSAGE_NO_SUCH_JOB_PLUGIN = "Failed to find job_plugin of type %s, available types include %s"


class CliInterface:
    """
    High-level interface for loading shell and job plugins and matching
    them to specified parameters.
    """

    def __init__(self):
        """ """
        module_prefix = self.__module__
        self.cli_shells = plugins_dict(f"{module_prefix}.shell", "__name__")
        self.cli_job_interfaces = plugins_dict(f"{module_prefix}.job", "__name__")
        self.active_cli_shells = {}

    def get_plugins(self, shell_params, job_params):
        """
        Return shell and job interface defined by and configured via
        specified params.
        """
        shell = self.get_shell_plugin(shell_params)
        job_interface = self.get_job_interface(job_params)
        return shell, job_interface

    def get_shell_plugin(self, shell_params):
        shell_plugin = shell_params.get("plugin", DEFAULT_SHELL_PLUGIN)
        requested_shell_settings = json.dumps(shell_params, sort_keys=True)
        if requested_shell_settings not in self.active_cli_shells:
            shell_plugin_class = self.cli_shells.get(shell_plugin)
            if not shell_plugin_class:
                raise ValueError(
                    f"Unknown shell_plugin [{shell_plugin}], available plugins are {list(self.cli_shells.keys())}"
                )
            self.active_cli_shells[requested_shell_settings] = shell_plugin_class(**shell_params)
        return self.active_cli_shells[requested_shell_settings]

    def get_job_interface(self, job_params):
        job_plugin = job_params.get("plugin")
        if not job_plugin:
            raise ValueError(ERROR_MESSAGE_NO_JOB_PLUGIN)
        job_plugin_class = self.cli_job_interfaces.get(job_plugin)
        if not job_plugin_class:
            raise ValueError(ERROR_MESSAGE_NO_SUCH_JOB_PLUGIN % (job_plugin, list(self.cli_job_interfaces.keys())))
        return job_plugin_class(**job_params)


def split_params(params):
    shell_params = {k.replace("shell_", "", 1): v for k, v in params.items() if k.startswith("shell_")}
    job_params = {k.replace("job_", "", 1): v for k, v in params.items() if k.startswith("job_")}
    return shell_params, job_params



from abc import ABC

import inspect


class UserAuthManager(ABC):
    """
    Authorization/Authentication manager.
    """

    def __init__(self, config):
        self._authorization_methods = []
        self._authentication_methods = []

        try:
            user_auth = config.get("user_auth", None)
            if not user_auth:
                return
            authentications = user_auth.pop("authentication", [])
            authorizations = user_auth.pop("authorization", [])

            for authorization in authorizations:
                authorization.update(user_auth)
                obj = get_object("pulsar.user_auth.methods." + authorization["type"], "auth_type",
                                 authorization["type"])
                self._authorization_methods.append(obj(authorization))

            for authentication in authentications:
                authentication.update(user_auth)
                obj = get_object("pulsar.user_auth.methods." + authentication["type"], "auth_type",
                                 authentication["type"])
                self._authentication_methods.append(obj(authentication))
        except Exception as e:
            raise Exception("cannot read auth configuration") from e

    def authorize(self, job_id, job_directory):
        authentication_info = self.__authenticate(job_id, job_directory)

        if len(self._authorization_methods) == 0:
            return True
        for method in self._authorization_methods:
            res = method.authorize(authentication_info)
            if res:
                return True

        raise Exception("Could not authorize job execution on remote resource")

    def __authenticate(self, job_id, job_directory):
        if len(self._authentication_methods) == 0:
            return {}
        for method in self._authentication_methods:
            res = method.authenticate(job_directory)
            if res:
                return res

        raise Exception("Could not authenticate job %s" % job_id)


def get_object(module_name, attribute_name, attribute_value):
    module = __import__(module_name)
    for comp in module_name.split(".")[1:]:
        module = getattr(module, comp)
    for _, obj in inspect.getmembers(module):
        if inspect.isclass(obj) and hasattr(obj, attribute_name) and getattr(obj, attribute_name) == attribute_value:
            return obj
    raise Exception("Cannot find object %s with attribute %s=%s " % (module_name, attribute_name, attribute_value))



import time

from .util import json_loads

MAX_RETRY_COUNT = 5
RETRY_SLEEP_TIME = 0.1


class parseJson:

    def __call__(self, func):
        def replacement(*args, **kwargs):
            response = func(*args, **kwargs)
            return json_loads(response)
        return replacement


class retry:

    def __call__(self, func):

        def replacement(*args, **kwargs):
            max_count = MAX_RETRY_COUNT
            count = 0
            while True:
                count += 1
                try:
                    return func(*args, **kwargs)
                except Exception:
                    if count >= max_count:
                        raise
                    else:
                        time.sleep(RETRY_SLEEP_TIME)
                        continue

        return replacement



from abc import ABC, abstractmethod


class AuthMethod(ABC):
    """
    Defines the interface to various authentication/authorization methods.
    """

    @abstractmethod
    def authorize(self, authentication_info):
        raise NotImplementedError("a concrete class should implement this")

    @abstractmethod
    def authenticate(self, job_directory):
        raise NotImplementedError("a concrete class should implement this")



try:
    from galaxy.jobs.runners.util.cli import (
        CliInterface,
        split_params,
    )
except ImportError:
    from pulsar.managers.util.cli import (  # type: ignore[no-redef]
        CliInterface,
        split_params,
    )


def build_cli_interface():
    return CliInterface()


def get_shell(params):
    cli_interface = build_cli_interface()
    shell_params, _ = split_params(params)
    return cli_interface.get_shell_plugin(shell_params)



"""Pulsar server-side integration with pulsar-relay.

This module provides functionality to bind Pulsar job managers to the
pulsar-relay, allowing them to receive control messages (setup, status
requests, kill) and publish status updates.
"""
import functools
import logging
import threading
import time

from pulsar import manager_endpoint_util
from pulsar.client.transport.relay import RelayTransport
from .relay_state import RelayState

log = logging.getLogger(__name__)


def bind_manager_to_relay(manager, relay_state: RelayState, relay_url, conf):
    """Bind a specific manager to the relay.

    Args:
        manager: Pulsar job manager instance
        relay_state: RelayState for managing consumer threads
        relay_url: URL of the pulsar-relay server
        conf: Configuration dictionary with relay credentials
    """
    manager_name = manager.name
    log.info("bind_manager_to_relay called for relay [%s] and manager [%s]", relay_url, manager_name)

    # Extract relay credentials
    username = conf.get('message_queue_username', 'admin')
    password = conf.get('message_queue_password')
    if not password:
        raise Exception("message_queue_password is required for relay communication")

    # Extract optional relay topic prefix
    relay_topic_prefix = conf.get('relay_topic_prefix', '')

    # Create relay transport
    relay_transport = RelayTransport(relay_url, username, password)

    # Define message handlers
    process_setup_messages = functools.partial(__process_setup_message, manager)
    process_kill_messages = functools.partial(__process_kill_message, manager)
    process_status_messages = functools.partial(__process_status_message, manager)

    # Determine topics based on manager name and optional prefix
    setup_topic = __make_topic_name(relay_topic_prefix, "job_setup", manager_name)
    status_request_topic = __make_topic_name(relay_topic_prefix, "job_status_request", manager_name)
    kill_topic = __make_topic_name(relay_topic_prefix, "job_kill", manager_name)
    status_update_topic = __make_topic_name(relay_topic_prefix, "job_status_update", manager_name)

    # Start consumer threads if message_queue_consume is enabled
    if conf.get("message_queue_consume", True):
        log.info("Starting relay consumer threads for manager '%s'", manager_name)

        # Single consumer thread for all control messages
        consumer_thread = start_consumer(
            relay_transport,
            relay_state,
            [setup_topic, status_request_topic, kill_topic],
            {
                setup_topic: process_setup_messages,
                status_request_topic: process_status_messages,
                kill_topic: process_kill_messages,
            }
        )

        relay_state.threads.append(consumer_thread)

    # Bind status change callback to publish status updates to relay
    if conf.get("message_queue_publish", True):
        log.info("Binding status change callback for manager '%s'", manager_name)

        def bind_on_status_change(new_status, job_id):
            job_id = job_id or 'unknown'
            try:
                message = "Publishing Pulsar state change with status %s for job_id %s via relay"
                log.debug(message, new_status, job_id)
                payload = manager_endpoint_util.full_status(manager, new_status, job_id)
                relay_transport.post_message(status_update_topic, payload)
            except Exception:
                log.exception("Failure to publish Pulsar state change for job_id %s via relay." % job_id)
                raise

        manager.set_state_change_callback(bind_on_status_change)


def start_consumer(relay_transport, relay_state: RelayState, topics, handlers):
    """Start a consumer thread that polls for messages.

    Args:
        relay_transport: RelayTransport instance
        relay_state: RelayState for checking if consumer should continue
        topics: List of topics to subscribe to
        handlers: Dict mapping topics to handler functions

    Returns:
        Thread object
    """
    def consume():
        log.info("Starting relay consumer for topics: %s", topics)

        while relay_state.active:
            try:
                # Long poll for messages (30 second timeout)
                messages = relay_transport.long_poll(topics, timeout=30)

                for message in messages:
                    topic = message.get('topic')
                    payload = message.get('payload', {})

                    handler = handlers.get(topic)
                    if handler:
                        try:
                            handler(payload)
                        except Exception:
                            job_id = payload.get('job_id', 'unknown')
                            log.exception("Failed to process message for job_id %s from topic %s", job_id, topic)
                    else:
                        log.warning("No handler found for topic '%s'", topic)

            except Exception:
                if relay_state.active:
                    log.exception("Exception while polling relay, will retry after delay.")
                    # Brief sleep before retrying
                    time.sleep(5)
                else:
                    log.debug("Exception during shutdown, stopping consumer.")
                    break

        log.info("Finished consuming relay messages - no more messages will be processed.")

    thread = threading.Thread(
        name="relay-consumer-%s" % "-".join(topics),
        target=consume
    )
    thread.daemon = True
    thread.start()
    return thread


def __process_setup_message(manager, body):
    """Process a job setup message.

    Args:
        manager: Job manager instance
        body: Message payload containing job setup parameters
    """
    job_id = __client_job_id_from_body(body)
    if not job_id:
        log.error('Could not parse job id from body: %s', body)
        return

    try:
        log.info("Processing setup message for job_id %s", job_id)
        manager_endpoint_util.submit_job(manager, body)
    except Exception:
        log.exception("Failed to process setup message for job_id %s", job_id)


def __process_status_message(manager, body):
    """Process a status request message.

    Args:
        manager: Job manager instance
        body: Message payload containing job_id
    """
    job_id = __client_job_id_from_body(body)
    if not job_id:
        log.error('Could not parse job id from body: %s', body)
        return

    try:
        log.debug("Processing status request for job_id %s", job_id)
        manager.trigger_state_change_callback(job_id)
    except Exception:
        log.exception("Failed to process status message for job_id %s", job_id)


def __process_kill_message(manager, body):
    """Process a job kill message.

    Args:
        manager: Job manager instance
        body: Message payload containing job_id
    """
    job_id = __client_job_id_from_body(body)
    if not job_id:
        log.error('Could not parse job id from body: %s', body)
        return

    try:
        log.info("Processing kill request for job_id %s", job_id)
        manager.kill(job_id)
    except Exception:
        log.exception("Failed to process kill message for job_id %s", job_id)


def __client_job_id_from_body(body):
    """Extract job_id from message body.

    Args:
        body: Message payload dictionary

    Returns:
        job_id string or None if not found
    """
    job_id = body.get("job_id", None)
    return job_id


def __make_topic_name(prefix, base_topic, manager_name):
    """Create a topic name with optional prefix and manager suffix.

    Args:
        prefix: Optional prefix string (e.g., 'galaxy1', 'prod')
        base_topic: Base topic name (e.g., 'job_setup', 'job_status_update')
        manager_name: Manager name (e.g., '_default_', 'cluster_a')

    Returns:
        Fully qualified topic name

    Examples:
        __make_topic_name('', 'job_setup', '_default_') -> 'job_setup'
        __make_topic_name('', 'job_setup', 'cluster_a') -> 'job_setup_cluster_a'
        __make_topic_name('prod', 'job_setup', '_default_') -> 'prod_job_setup'
        __make_topic_name('prod', 'job_setup', 'cluster_a') -> 'prod_job_setup_cluster_a'
    """
    parts = []

    # Add prefix if provided
    if prefix:
        parts.append(prefix)

    # Add base topic
    parts.append(base_topic)

    # Add manager name if not default
    if manager_name != "_default_":
        parts.append(manager_name)

    return "_".join(parts)



import json
from logging import getLogger

from .torque import Torque

log = getLogger(__name__)


class OpenPBS(Torque):

    ERROR_MESSAGE_UNRECOGNIZED_ARG = "Unrecognized long argument passed to OpenPBS CLI plugin: %s"

    def get_status(self, job_ids=None):
        return "qstat -f -F json"

    def get_single_status(self, job_id):
        return f"qstat -f {job_id}"

    def parse_status(self, status, job_ids):
        try:
            data = json.loads(status)
        except Exception:
            log.warning(f"No valid qstat JSON return from `qstat -f -F json`, got the following: {status}")
        rval = {}
        for job_id, job in data.get("Jobs", {}).items():
            if job_id in job_ids:
                # map PBS job states to Galaxy job states.
                rval[id] = self._get_job_state(job["job_state"])
        return rval


__all__ = ("OpenPBS",)
